"""
    Info print
    ~~~~~~~~~~

    Insert in every stdout.write() a info line from witch code line this print comes.
    Usefull to find debug print statements ;)

    WARNING: This is very slow and should only be used with the developer server ;)

    Simply put this two lines in your settings:
        ----------------------------------------------------------------------
        from django_tools.utils import info_print
        info_print.redirect_stdout()
        ----------------------------------------------------------------------

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author:$

    :copyleft: 2009-2010 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import inspect
import os
import sys
import warnings


MAX_FILEPATH_LEN = 66


class InfoStdout:
    """ Insert in every stdout.write() a info line from witch code line this print comes."""

    def __init__(self, orig_stdout):
        self.orig_stdout = orig_stdout
        self.old_fileinfo = None

    def write(self, txt):
        fileinfo = self._get_fileinfo()
        if fileinfo != self.old_fileinfo:
            self.orig_stdout.write(f"\n{fileinfo}:\n{txt}")
            self.old_fileinfo = fileinfo
        else:
            self.orig_stdout.write(txt)

    def flush(self):
        self.orig_stdout.flush()

    def _get_fileinfo(self):
        """ return fileinfo: Where from the announcement comes? """
        try:
            self_basename = os.path.basename(__file__)
            if self_basename.endswith(".pyc"):
                # cut: ".pyc" -> ".py"
                self_basename = self_basename[:-1]

            for stack_frame in inspect.stack():
                # go forward in the stack, to outside of this file.
                filename = stack_frame[1]
                lineno = stack_frame[2]
                if os.path.basename(filename) != self_basename:
                    break

            if len(filename) >= MAX_FILEPATH_LEN:
                filename = f"...{filename[-MAX_FILEPATH_LEN:]}"
            fileinfo = f"{filename} line {lineno}"
        except Exception as e:
            fileinfo = f"(inspect Error: {e})"

        return fileinfo


__redirected = False


def redirect_stdout():
    global __redirected

    if not __redirected:
        __redirected = True
        try:
            warnings.warn("Redirect stdout/stderr for info print!")
            orig_stdout = sys.stdout
            sys.stdout = InfoStdout(orig_stdout)
            orig_stderr = sys.stderr
            sys.stderr = InfoStdout(orig_stderr)
        except Exception as err:
            print("Error:", err)



"""
    need full validators
    ~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2010-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import os
import re
from urllib.parse import urlsplit

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator, URLValidator
from django.utils.translation import gettext_lazy as _


# re from django.utils.translation.trans_real.accept_language_re
language_code_re = re.compile(r'^([A-Za-z]{1,8}(?:-[A-Za-z]{1,8})*)$')
validate_language_code = RegexValidator(
    language_code_re,
    _('Enter a valid language code (Accept-Language header format, see RFC2616)'),
    'invalid'
)


class ExistingDirValidator:
    def __init__(self, base_path=settings.MEDIA_ROOT):
        self.base_path = os.path.normpath(os.path.abspath(base_path))

    def __call__(self, value):
        abs_path = os.path.normpath(os.path.abspath(os.path.join(self.base_path, value)))

        if not abs_path.startswith(self.base_path):
            if settings.DEBUG:
                msg = _(f"Directory {abs_path!r} is not in base path ('{self.base_path}')")
            else:
                msg = _("Directory is not in base path!")
            raise ValidationError(msg)

        if not os.path.isdir(abs_path):
            if settings.DEBUG:
                msg = _("Directory %r doesn't exist!") % abs_path
            else:
                msg = _("Directory doesn't exist!")

            raise ValidationError(msg)

        return abs_path


class URLValidator2(URLValidator):
    """
    A flexible version of the original django URLValidator ;)

    scheme://netloc/path?query#fragment
    """
    regex = re.compile(r'^.+$', re.IGNORECASE)

    def __init__(
            self,
            allow_schemes=(
                "http",
                "https"),
            allow_all_schemes=False,
            allow_netloc=True,
            allow_query=True,
            allow_fragment=True):
        super().__init__()

        if __debug__ and (allow_schemes or allow_all_schemes) and not allow_netloc:
            raise AssertionError("Can't allow schemes without netloc!")

        if __debug__ and allow_schemes and allow_all_schemes:
            raise Warning("allow_schemes would be ignored, while allow_all_schemes==True!")

        self.allow_schemes = allow_schemes or ()
        self.allow_all_schemes = allow_all_schemes
        self.allow_netloc = allow_netloc
        self.allow_query = allow_query
        self.allow_fragment = allow_fragment

    def __call__(self, value):
        scheme, netloc, path, query, fragment = urlsplit(value)

        if (scheme or netloc) and \
                not self.allow_schemes and \
                not self.allow_all_schemes and \
                not self.allow_netloc:
            raise ValidationError(
                _("Please enter a local URL (without protocol/domain)."), code="local"
            )

        if scheme:
            if not self.allow_all_schemes and scheme not in self.allow_schemes:
                raise ValidationError(_("The URL doesn't start with a allowed scheme."), "scheme")

        if netloc and not self.allow_netloc:
            raise ValidationError(_('Enter a valid URL without domain.'), code='netloc')

        if query and not self.allow_query:
            raise ValidationError(_('Enter a valid URL without a query.'), code='query')

        if fragment and not self.allow_fragment:
            raise ValidationError(_('Enter a valid URL without a fragment.'), code='fragment')



"""
    :created: 21.08.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging
import time

from django.conf import settings
from django.contrib import messages
from django.core.cache import cache


log = logging.getLogger(__name__)


class DelayBase:
    """
    Delay something via time.sleep() and used delay value from GET parameters.

    This is only the base class, use:
        - SessionDelay()
    or
        - CacheDelay()
    """
    backend_name = None

    def __init__(self, key, only_debug=True):
        """
        :param key: session/cache key for store the delay value into backend
        :param only_debug: If True: delay only if settings.DEBUG == True
        """
        self.key = key
        self.only_debug = only_debug

    def _delete(self):
        raise NotImplementedError

    def _set(self, sec):
        raise NotImplementedError

    def _get(self):
        raise NotImplementedError

    def load(self, request, query_string, default=5):
        """
        :param query_string: the request.GET key that activate/set delay value
        :param default: fallback if no value in GET
        """
        if query_string not in request.GET:
            if self._get():
                log.debug("Delete %r delay from %s", self.key, self.backend_name)
            self._delete()
            return

        if self.only_debug and not settings.DEBUG:
            log.debug("Ignore ?%s, because DEBUG is not ON!", query_string)
            self._delete()
            return

        log.info("Add '%s' value to %s", query_string, self.backend_name)
        sec = request.GET[query_string]
        if not sec:
            sec = default
        else:
            try:
                sec = int(sec)
            except ValueError:
                sec = float(sec)
        log.warning("Save %s sec. from %r for %r into %s", sec, query_string, self.key, self.backend_name)

        if request.user.is_authenticated:
            messages.warning(request, f"Use {sec} sec. {query_string!r}")

        self._set(sec)

    def sleep(self):
        sec = self._get()

        if sec is None:
            log.debug("No delay for %r from %s", self.key, self.backend_name)
            return

        log.warning("Delay %s sec. for %r", sec, self.key)
        time.sleep(sec)


class SessionDelay(DelayBase):
    """
    Delay via time.sleep() something, use the sleep seconds from GET parameter.
    Store the sleep seconds into request.session
    """
    backend_name = "session"

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super().__init__(*args, **kwargs)

    def _delete(self):
        try:
            del self.request.session[self.key]
        except KeyError:
            pass

    def _set(self, sec):
        self.request.session[self.key] = sec

    def _get(self):
        return self.request.session.get(self.key)


class CacheDelay(DelayBase):
    """
    Delay via time.sleep() something, use the sleep seconds from GET parameter.
    Store the sleep seconds into django cache backend
    """
    backend_name = "cache"

    def __init__(self, *args, cache_timeout=30 * 60, **kwargs):
        """
        :param cache_timeout: timeout, in seconds, to use for the cache.

        You can set TIMEOUT to None so that, by default, cache keys never expire.
        see also:
            https://docs.djangoproject.com/en/1.11/topics/cache/#cache-arguments
        """
        super().__init__(*args, **kwargs)
        self.cache_timeout = cache_timeout

    def _delete(self):
        cache.delete(self.key)

    def _set(self, sec):
        cache.set(self.key, sec, self.cache_timeout)

    def _get(self):
        return cache.get(self.key)



"""
    need full model and form fields
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2010-2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django import forms
from django.db import models
from django.utils.translation import gettext_lazy as _

from django_tools import validators


class LanguageCodeFormField(forms.CharField):
    """
    Language Code form field in Accept-Language header format (RFC 2616)

    >>> LanguageCodeFormField().clean('en')
    'en'

    >>> LanguageCodeFormField().clean('en-GB')
    'en-GB'

    >>> try:
    ...     LanguageCodeFormField().clean("this is wrong")
    ... except Exception as err:
    ...     print(err.__class__.__name__, err)
    ValidationError ['Enter a valid language code (Accept-Language header format, see RFC2616)']

    >>> try:
    ...     LanguageCodeFormField().clean(None)
    ... except Exception as err:
    ...     print(err.__class__.__name__, err)
    ValidationError ['This field is required.']

    >>> LanguageCodeFormField(required=False).clean(None)
    ''
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.append(validators.validate_language_code)


class LanguageCodeModelField(models.CharField):
    """
    >>> LanguageCodeModelField(max_length=20).run_validators('en-GB')

    >>> try:
    ...     LanguageCodeModelField(max_length=20).run_validators("this is wrong")
    ... except Exception as err:
    ...     print(err.__class__.__name__, err)
    ValidationError ['Enter a valid language code (Accept-Language header format, see RFC2616)']
    """
    default_validators = [validators.validate_language_code]
    description = _("Language Code in Accept-Language header format defined in RFC 2616")



"""
    ThrottledAdminEmailHandler
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    created 25.09.2019 by Jens Diemer
    :copyleft: 2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging
import time

from django.core import mail
from django.utils.log import AdminEmailHandler


log = logging.getLogger(__name__)


class ThrottledAdminEmailHandler(AdminEmailHandler):
    """
    Throttled email handler.

    Works similar as the origin django.utils.log.AdminEmailHandler
    But mails send in >min_delay_sec< time range will be skiped.
    The mail subject of skipped mails will be added to the next error mails.

    Note: Currently "include_html" is not supported and will be deactivated!

    usage, e.g.:

        LOGGING = {
            # ...
            "handlers": {
                "mail_admins": {
                    "level": "ERROR",
                    "class": "django_tools.log_utils.throttle_admin_email_handler.ThrottledAdminEmailHandler",
                    "formatter": "email",
                    "min_delay_sec": 20, # << -- skip mails in this time range
                },
                # ...
            },
            # ...
        }
    """

    def __init__(self, *args, min_delay_sec=30, **kwargs):
        super().__init__(*args, **kwargs)

        self.include_html = False  # TODO: support html mails, too

        self.min_delay_sec = min_delay_sec
        self.next_mail = None
        self.skipped_subjects = []

    def send_mail(self, subject, message, *args, **kwargs):
        if self.next_mail:
            if self.next_mail > time.time():
                log.debug("Throttle error mail: %r", subject)
                self.skipped_subjects.append(subject)
                return

        if self.skipped_subjects:
            prefix = "\nNote: there are %i skipped mails:\n" % len(self.skipped_subjects)
            prefix += "\n\t* " + "\n\t* ".join(self.skipped_subjects)
            prefix += "\n\n"
            message = prefix + message
            self.skipped_subjects.clear()

        log.debug("Send mail: %r", subject)

        mail.mail_admins(subject, message, *args, connection=self.connection(), **kwargs)

        self.next_mail = time.time() + self.min_delay_sec



# Generated by Django 3.1.4 on 2020-12-06 08:23
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import django_tools.serve_media_app.models


def forward_code(apps, schema_editor):
    app_name, model_name = settings.AUTH_USER_MODEL.split('.')
    User = apps.get_model(app_name, model_name)
    UserMediaTokenModel = apps.get_model('serve_media_app', 'UserMediaTokenModel')
    for user in User.objects.all():
        UserMediaTokenModel.objects.get_or_create(user_id=user.pk)


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserMediaTokenModel',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('token',
                 models.CharField(
                     default=django_tools.serve_media_app.models.generate_token,
                     max_length=12,
                     unique=True)),
                ('user',
                 models.ForeignKey(
                     editable=False,
                     on_delete=django.db.models.deletion.CASCADE,
                     related_name='token',
                     to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.RunPython(
            forward_code,
            reverse_code=migrations.RunPython.noop),
    ]



"""
    $ ./manage.py nice_diffsettings

    The code based on:
        django.core.management.commands.diffsettings
"""

import pprint

from django.core.management.base import BaseCommand


def module_to_dict(module, omittable=lambda k: k.startswith("_")):
    """
    Same as django.core.management.commands.diffsettings.module_to_dict
    but didn't use repr() ;)
    """
    return {k: v for k, v in module.__dict__.items() if not omittable(k)}


class Command(BaseCommand):
    help = """Displays differences between the current settings.py and Django's
    default settings in a pretty-printed representation."""

    requires_system_checks = False

    def add_arguments(self, parser):
        parser.add_argument(
            "--all",
            action="store_true",
            dest="all",
            default=False,
            help="Display all settings, regardless of their value.",
        )

    def handle(self, **options):
        from django.conf import global_settings, settings

        # Because settings are imported lazily, we need to explicitly load them.
        settings._setup()

        user_settings = module_to_dict(settings._wrapped)
        default_settings = module_to_dict(global_settings)

        self.stdout.write("-" * 79)

        for key in sorted(user_settings):
            display = False
            if key not in default_settings:
                display = True
            elif user_settings[key] != default_settings[key]:
                display = True
            elif options["all"]:
                display = True

            if display:

                value = user_settings[key]
                try:
                    pformated = pprint.pformat(value)
                except Exception as err:
                    # e.g.: https://github.com/andymccurdy/redis-py/issues/995
                    pformated = f"<Error: {err}>"

                self.stdout.write(f"{key} = {pformated}\n\n")

        self.stdout.write("-" * 79)



"""

    TODO: Move this!
    from: django_tools/unittest_utils/logging_utils.py
    to..: django_tools/logging_utils.py

    :created: 2015 by Jens Diemer
    :copyleft: 2015-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging


class CutPathnameLogRecordFactory:
    """
    Adds 'cut_path' attribute on log record. So '%(cut_path)s' can be used in log formatter.

    Add this to you settings.py, e.g.:

        import logging
        from django_tools.unittest_utils.logging_utils import CutPathnameLogRecordFactory

        logging.setLogRecordFactory(CutPathnameLogRecordFactory(max_length=50))

        LOGGING = {
            # ...
            'formatters': {
                'verbose': {
                    'format': '%(levelname)8s %(cut_path)s:%(lineno)-3s %(message)s'
                },
            },
            # ...
        }
    """

    def __init__(self, max_length=40):
        self.max_length = max_length
        self.origin_factory = logging.getLogRecordFactory()

    def cut_path(self, pathname):
        if len(pathname) <= self.max_length:
            return pathname
        return f"...{pathname[-(self.max_length - 3):]}"

    def __call__(self, *args, **kwargs):
        record = self.origin_factory(*args, **kwargs)
        record.cut_path = self.cut_path(record.pathname)
        return record


class FilterAndLogWarnings:
    """
    Filter warnings and pipe them to logging system
    Warnings of external packages are displayed only once and only the file path.

    Activate warnings, e.g.:
        export PYTHONWARNINGS=all

    Add this to you settings.py, e.g.:

        import warnings
        import logging
        from django_tools.unittest_utils.logging_utils import FilterAndLogWarnings

        warnings.showwarning = FilterAndLogWarnings()
    """

    skipped_filenames = []

    def __init__(self, logger_name=None, external_package_paths=None):
        if logger_name is None:
            logger_name = __name__
        self.logger = logging.getLogger(logger_name)

        if external_package_paths is None:
            self.external_package_paths = ("/dist-packages/", "/site-packages/")
        else:
            self.external_package_paths = external_package_paths

    def __call__(self, message, category, filename, lineno, args=None, exc_info=None):

        # handle warnings from external packages:
        for path_part in self.external_package_paths:
            if path_part in filename:
                if filename not in self.skipped_filenames:
                    self.skipped_filenames.append(filename)
                    self.logger.warning(f"There are warnings in: {filename}")
                return

        # Create log entry for the warning:
        record = self.logger.makeRecord(
            name=self.logger.name,
            level=logging.WARNING,
            fn=filename,
            lno=lineno,
            msg=f"{category.__name__}:{message}",
            args=args,
            exc_info=exc_info,
        )
        self.logger.handle(record)



"""
Print the query log to standard out.

Useful for optimizing database calls.

Inspired by the method at: <http://www.djangosnippets.org/snippets/344/>
"""
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object  # fallback for Django < 1.10


class QueryLogMiddleware(MiddlewareMixin):

    def process_response(self, request, response):
        from django.conf import settings
        from django.db import connection

        if settings.DEBUG:
            queries = {}
            for query in connection.queries:
                sql = query["sql"]
                queries.setdefault(sql, 0)
                queries[sql] += 1
            duplicates = sum([count - 1 for count in list(queries.values())])
            print("------------------------------------------------------")
            print(f"Total Queries:     {len(queries)}")
            print(f"Duplicate Queries: {duplicates}")
            print()
            for query, count in list(queries.items()):
                print(f"{count} x {query}")
            print("------------------------------------------------------")
        return response



"""
    filemanager
    ~~~~~~~~~~~

    Stuff to build a app like a file manager.


    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import datetime
import grp
import os
import pwd
import stat
from operator import attrgetter

from django.contrib import messages

from django_tools.filemanager.exceptions import FilemanagerError
from django_tools.filemanager.filesystem_browser import BaseFilesystemBrowser
from django_tools.filemanager.utils import symbolic_notation


class BaseFilesystemObject:
    def __init__(self, base_path, name, abs_path, link_path=None):
        self.base_path = base_path  # path in which this item exists
        self.name = name  # The name of the directory item
        self.abs_path = abs_path  # absolute path to this dir item
        self.link_path = link_path  # Only for links: the real path of the dir item

        self.stat = os.stat(self.abs_path)
        self.size = self.stat[stat.ST_SIZE]
        self.mode = self.stat[stat.ST_MODE]
        self.mtime = datetime.datetime.fromtimestamp(self.stat[stat.ST_MTIME])

        self.mode_octal = oct(self.mode)
        self.mode_symbol = symbolic_notation(self.mode)

        self.uid = self.stat[stat.ST_UID]
        self.username = pwd.getpwuid(self.uid).pw_name
        self.gid = self.stat[stat.ST_GID]
        self.groupname = grp.getgrgid(self.gid).gr_name

    def __repr__(self):
        return f"{self.item_type} '{self.name}' in {self.base_path}"


class BaseFileItem(BaseFilesystemObject):
    is_file = True
    is_dir = False
    item_type = "file"


class BaseDirItem(BaseFilesystemObject):
    is_file = False
    is_dir = True
    item_type = "dir"


class BaseFileLinkItem(BaseFileItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.item_type = f"file link to {self.link_path}"


class BaseDirLinkItem(BaseDirItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.item_type = f"dir link to {self.link_path}"


# ------------------------------------------------------------------------------


class BaseFilemanager(BaseFilesystemBrowser):
    """
    Base class for building a filemanager
    """
    DIR_ITEM = BaseDirItem
    FILE_ITEM = BaseFileItem
    DIR_LINK_ITEM = BaseDirLinkItem
    FILE_LINK_ITEM = BaseFileLinkItem

    def __init__(self, request, absolute_path, base_url, rest_url, allow_upload=False):
        super().__init__(request, absolute_path, base_url, rest_url)

        self.allow_upload = allow_upload
        self.dir_items = self.read_dir(self.absolute_path)

    def read_dir(self, path):
        dir_items = []
        for item in os.listdir(path):
            item_abs_path = os.path.join(self.absolute_path, item)
            link_path = None
            if os.path.islink(item_abs_path):
                link_path = os.readlink(item_abs_path)
                if os.path.isdir(link_path):
                    item_class = self.DIR_LINK_ITEM
                elif os.path.isfile(link_path):
                    item_class = self.FILE_LINK_ITEM
                else:
                    raise NotImplementedError
            elif os.path.isdir(item_abs_path):
                item_class = self.DIR_ITEM
            elif os.path.isfile(item_abs_path):
                item_class = self.FILE_ITEM
            else:
                messages.info(self.request, f"unhandled directory item: {self.absolute_path!r}")
                continue

            instance = self.get_filesystem_item_instance(item_class, item, item_abs_path, link_path)
            dir_items.append(instance)

        # sort the dir items by name but directories first
        # http://wiki.python.org/moin/HowTo/Sorting/#Operator_Module_Functions
        dir_items = sorted(dir_items, key=attrgetter('item_type', 'name'))

        return dir_items

    def get_filesystem_item_instance(self, item_class, item, item_abs_path, link_path):
        """
        Good point for overwrite, to add attributes to the filesystem items.
        """
        instance = item_class(self.absolute_path, item, item_abs_path, link_path)
        return instance

    def handle_uploaded_file(self, f):
        if not self.allow_upload:
            raise FilemanagerError("Upload not allowed here!")

        path = os.path.join(self.absolute_path, f.name)
        destination = open(path, 'wb+')
        for chunk in f.chunks():
            destination.write(chunk)
        destination.close()

        messages.success(self.request,
                         f"File '{f.name}' ({f.size:d} Bytes) uploaded to {self.absolute_path}"
                         )



"""
    permission helpers
    ~~~~~~~~~~~~~~~~~~

    create 03.03.2017 by Jens Diemer
"""


import logging

from django.conf import settings
from django.contrib.admin.sites import site
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.db import NotSupportedError


log = logging.getLogger(__name__)


def pformat_permission(permission):
    """
    Generate the permission string in the same format as user.has_perm() argument:
        "<appname>.<codename>"
    """
    assert isinstance(permission, Permission), "No auth.models.Permission instance!"
    return f"{permission.content_type.app_label}.{permission.codename}"


def get_permission(app_label, codename):
    try:
        perm_obj = Permission.objects.all().get(
            content_type__app_label=app_label,
            codename=codename
        )
    except Permission.DoesNotExist as err:
        log.exception("Error get permission '%s.%s':%s", app_label, codename, err)

        content_types = ContentType.objects.all().filter(app_label=app_label)
        if content_types.count() == 0:
            qs = ContentType.objects.all().values_list("app_label", flat=True).order_by("app_label")
            try:
                app_lables = ", ".join(qs.distinct("app_label"))
            except NotSupportedError:
                # e.g.: sqlite has no distinct
                app_lables = ", ".join(set(qs))
            raise PermissionDenied(
                (
                    "App label '%s' from permission '%s.%s' doesn't exists!"
                    " All existing labels are: %s"
                ) % (app_label, app_label, codename, app_lables)
            )
        qs = Permission.objects.all().filter(content_type__in=content_types)
        if qs.filter(codename=codename).count() == 0:
            codenames = qs.values_list("codename", flat=True).order_by("codename")
            raise PermissionDenied(
                (
                    "Codename '%s' from permission '%s.%s' doesn't exists!"
                    " All existing codenames are: %s"
                ) % (codename, app_label, codename, ", ".join(codenames))
            )
    else:
        return perm_obj


def get_permission_by_string(permission):
    try:
        app_label, codename = permission.split(".")
    except ValueError as err:
        raise PermissionDenied(
            f"Wrong permission string format '{permission}': {err}"
        )
    return get_permission(app_label, codename)


def add_permissions(permission_obj, permissions):
    """
    Add permissions

    e.g.:
        group = Group.objects.create(name="Foobar Group")
        permissions=(
            (YourModelClass, "the_permission_string"),
        )
        add_permissions(permission_obj=group, permissions=permissions)
    """
    for model_class, codename in permissions:
        content_type = ContentType.objects.get_for_model(model_class)
        app_label = content_type.app_label
        permission = get_permission(app_label, codename)
        permission_obj.permissions.add(permission)


def permissions2list(permissions):
    """
    e.g.:

        permissions = our_user_group.permissions.all()
        permissions = permissions2list(permissions)
        pprint.pprint(permissions)
        ['auth.user.add_user',
         'auth.user.change_user',
         'auth.user.delete_user']
    """
    permissions = [pformat_permission(permission) for permission in permissions]
    return permissions


def log_user_permissions(user, log_callable=None):
    """
    Helper to log all existing user permissions.

    e.g.:
        from django_tools.permissions import log_group_permissions
        log = logging.getLogger(__name__)

        log_user_permissions(the_user_instance, log_callable=log.debug)
    """
    if log_callable is None:
        log_callable = log.debug

    permissions = sorted(user.get_all_permissions())  # A string list!

    if not permissions:
        log_callable("User '%s' has no permission!", user.username)
    else:
        permissions = "\n".join(["* %s" % p for p in permissions])
        log_callable("User '%s' has permissions:\n%s", user.username, permissions)


def log_group_permissions(group, log_callable=None):
    """
    Helper to log all existing user group permissions.

    e.g.:
        from django_tools.permissions import log_group_permissions
        log = logging.getLogger(__name__)

        log_group_permissions(user_group_instance, log_callable=log.debug)
    """
    if log_callable is None:
        log_callable = log.debug

    permissions = group.permissions.all().order_by("content_type", "codename")
    if not permissions:
        log_callable("User group '%s' has no permission!", group.name)
    else:
        permissions = "\n".join([
            "* %s" % pformat_permission(permission)
            for permission in permissions
        ])
        log_callable("User group '%s' has permissions:\n%s", group.name, permissions)


def get_admin_permissions():
    """
    :return: list of Permission instances for models that are registered in the admin
    """
    # Create a content type list of all models that are registered in the admin
    admin_content_types = []
    for model in site._registry.keys():
        content_type = ContentType.objects.get_for_model(model)
        admin_content_types.append(content_type)

    # Add all permissions for all admin models:
    permissions = Permission.objects.all()
    permissions = permissions.filter(content_type__in=admin_content_types)
    return permissions


def add_app_permissions(permission_obj, app_label):
    """
    Add all permission from one django app to the given object.
    e.g.:
        add_app_permissions(permission_obj=user_group_instance, app_label="filer")
    """
    content_types = ContentType.objects.filter(app_label=app_label)
    permissions = Permission.objects.filter(content_type__in=content_types)
    log.debug(f"Add {permissions.count():d} permissions from app '{app_label}'")
    for permission in permissions:
        permission_obj.permissions.add(permission)


def check_permission(user, permission, raise_exception=True):
    """
    Check if given user has the given permission

    :param user: django user model instance
    :param permission: string e.g.: 'auth.user.add_user'
    :param raise_exception: raise PermissionDenied if user has not the permission
    :return: bool if user has permission
    """
    if user.is_superuser:
        log.debug("Don't check permissions for superuser, ok.")
        return True

    if user.has_perm(permission):
        log.debug('User "%s" has permission "%s"', user, permission)
        return True

    if settings.DEBUG:
        # 'validate' permission string
        # get_permission() will raise helpfull errors if format is wrong
        # or if the permission doesn't exists
        get_permission_by_string(permission)

    if raise_exception:
        log.error(
            'User "%s" has not permission "%s" -> raise PermissionDenied!',
            user, permission
        )
        raise PermissionDenied
    else:
        log.debug('User "%s" has not permission "%s"', user, permission)
        return False


def has_perm(user, permission):
    """
    Verbose check if user has permission:
    Create log entry if user has not the permission
    """
    if not user.has_perm(permission):
        log.debug('User %s has not %s', user, permission)
        return False
    return True


class ModelPermissionMixin:
    """
    Helper for easy model permission checks.
    e.g.:
        class FooModel(ModelPermissionMixin, models.Model):
            ...
            @classmethod
            def has_extra_permission_permission(cls, user, raise_exception=True):
                permission = cls.extra_permission_name(action="extra_permission")
                return check_permission(user, permission, raise_exception)

            class Meta:
                permissions = (
                    ("extra_permission", "Extra permission"),
                )

        def view(request):
            FooModel.has_change_permission(request.user, raise_exception=True)
            FooModel.has_extra_permission_permission(request.user, raise_exception=True)

    See also model used in our tests:
        django_tools_test_project.django_tools_test_app.models.PermissionTestModel
    """
    @classmethod
    def default_permission_name(cls, action):
        """
        Built the permission name for the default add/change/delete permissions.

        Note: Django will add "model name" to these permissions!
        e.g.:
            applabel.add_modelname
            applabel.change_modelname
            applabel.delete_modelname

        https://docs.djangoproject.com/en/1.8/ref/models/options/#default-permissions
        """
        assert action in cls._meta.default_permissions, f"'{action}' not in Meta.default_permissions !"
        permission = f"{cls._meta.app_label}.{action}_{cls._meta.model_name}"
        return permission

    @classmethod
    def extra_permission_name(cls, action):
        """
        Built permission name for own/extra permissions defined by "ModelClass.Meta.permissions"

        Note: Django will create these permissions without "model name"!
        e.g.:
            applabel.permission_name_modelname
        """
        if settings.DEBUG:
            all_permissions = [p[0] for p in cls._meta.permissions]
            assert action in all_permissions, (
                f"'{action}' not in Meta.permissions!"
                f" Existing keys are: {', '.join(all_permissions)}"
            )
        permission = f"{cls._meta.app_label}.{action}"
        return permission

    @classmethod
    def has_add_permission(cls, user, raise_exception=True):
        permission = cls.default_permission_name(action="add")
        return check_permission(user, permission, raise_exception)

    @classmethod
    def has_change_permission(cls, user, raise_exception=True):
        permission = cls.default_permission_name(action="change")
        return check_permission(user, permission, raise_exception)

    @classmethod
    def has_delete_permission(cls, user, raise_exception=True):
        permission = cls.default_permission_name(action="delete")
        return check_permission(user, permission, raise_exception)


def get_filtered_permissions(
    exclude_app_labels=None,
    exclude_actions=None,
    exclude_models=None,
    exclude_codenames=None,
    exclude_permissions=None
):
    """
    Generate a Permission instance list and exclude parts of it.

    Filters:
        <app_label>.<action>_<modelname>
                    `-----codename-----'

    usage, e.g.:
        permissions = get_filtered_permissions(
            exclude_app_labels=("easy_thumbnails", "filer"),
            exclude_actions=("delete",)
            exclude_models=(LimitToUsergroupsTestModel, PermissionTestModel),
            exclude_codenames=("add_group", "add_user"),
            exclude_permissions=(
                (ContentType, "add_contenttype"),
                (ContentType, "delete_contenttype"),
            )
        )
    """
    qs = Permission.objects.all().order_by("content_type__app_label", "content_type__model", "codename")

    if exclude_codenames is not None:
        qs = qs.exclude(codename__in=exclude_codenames)

    if exclude_actions is not None:
        for action in exclude_actions:
            codename_prefix = f"{action}_"
            qs = qs.exclude(codename__startswith=codename_prefix)

    if exclude_app_labels is not None:
        # SQLite doesn't support .distinct, so we can't do this:
        app_lables = set(ContentType.objects.all().values_list("app_label", flat=True))

        exclude_content_types = []
        for app_label in exclude_app_labels:
            if app_label not in app_lables:
                raise AssertionError(
                    f"app label {app_label!r} not found! Existing labels: {','.join(app_lables)}"
                )
            print(f"Check {app_label!r}")
            content_types = ContentType.objects.all().filter(app_label=app_label)
            assert content_types.count() > 0
            exclude_content_types += content_types

        qs = qs.exclude(content_type__in=exclude_content_types)

    if exclude_models is not None:
        for model in exclude_models:
            content_type = ContentType.objects.get_for_model(model)
            qs = qs.exclude(content_type=content_type)

    if exclude_permissions is not None:
        for model, codename in exclude_permissions:
            content_type = ContentType.objects.get_for_model(model)
            qs = qs.exclude(content_type=content_type, codename=codename)

    return qs


def pprint_filtered_permissions(permissions):
    """
    print a list like this:
        [*] auth.group.add_group
        [*] auth.group.change_group
        [ ] auth.group.delete_group
        [*] auth.permission.add_permission
        [ ] auth.permission.change_permission
        [*] auth.permission.delete_permission
        [*] auth.user.add_user
        [*] auth.user.change_user
        [ ] auth.user.delete_user
        ...
    """
    for permission in permissions:
        assert isinstance(permission, Permission), "List must contain auth.models.Permission instances!"

    qs = Permission.objects.all().order_by("content_type__app_label", "content_type__model", "codename")
    for permission in qs:
        contains = "[*]" if permission in permissions else "[ ]"
        print(f"{contains} {pformat_permission(permission)}")



"""
    per-site cache middleware
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    more information in the README.

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import logging
import sys

from django.conf import settings
from django.contrib import messages
from django.core.cache import cache
from django.http import HttpResponse
from django.utils.cache import get_max_age, patch_response_headers

from django_tools.utils.importlib import get_attr_from_settings


logger = logging.getLogger("django-tools.CacheMiddleware")

CACHE_MIDDLEWARE_ANONYMOUS_ONLY = getattr(settings, 'CACHE_MIDDLEWARE_ANONYMOUS_ONLY', False)
RUN_WITH_DEV_SERVER = getattr(settings, "RUN_WITH_DEV_SERVER", "runserver" in sys.argv)
EXTRA_DEBUG = getattr(settings, "CACHE_EXTRA_DEBUG", False)

COUNT_FETCH_FROM_CACHE = getattr(settings, "COUNT_FETCH_FROM_CACHE", False)
COUNT_UPDATE_CACHE = getattr(settings, "COUNT_UPDATE_CACHE", False)
COUNT_IN_CACHE = getattr(settings, "COUNT_IN_CACHE", False)

cache_callback = get_attr_from_settings("CACHE_CALLBACK", "DjangoTools cache callback")
logger.debug(f"Use cache callback: {cache_callback!r}")


CACHE_REQUESTS = "DJANGOTOOLS_CACHE_REQUESTS_COUNT"
CACHE_REQUEST_HITS = "DJANGOTOOLS_CACHE_REQUEST_HITS_COUNT"
CACHE_RESPONSES = "DJANGOTOOLS_CACHE_RESPONSES_COUNT"
CACHE_RESPONSE_HITS = "DJANGOTOOLS_CACHE_RESPONSE_HITS_COUNT"

_CACHE_KEYS = (CACHE_REQUESTS, CACHE_REQUEST_HITS, CACHE_RESPONSES, CACHE_RESPONSE_HITS)


# Some information from the cache usage.
# Note: These values are only for current process/thread valid and they not thread safe!
# Set settings.COUNT_IN_CACHE=True for process/thread safe counting.
LOCAL_CACHE_INFO = {
    # from FetchFromCacheMiddleware:
    "requests": None,  # total numbers of requests
    "request hits": None,  # number of cache hits

    # from UpdateCacheMiddleware:
    "responses": None,  # total numbers of responses
    "response hits": None,  # number of responses from cache
}


def init_cache_counting():
    """
    prepare LOCAL_CACHE_INFO and the counter in cache:
    All initial count value should be None, if count is disabled
    and should be 0, if enabled.
    """
    for key in _CACHE_KEYS:
        cache.delete(key)  # delete old entrie, if exist

    if COUNT_FETCH_FROM_CACHE:
        # Count in FetchFromCacheMiddleware is enabled
        LOCAL_CACHE_INFO["requests"] = 0
        LOCAL_CACHE_INFO["request hits"] = 0
        if COUNT_IN_CACHE:
            # Count in cache is enabled
            cache.set(CACHE_REQUESTS, 0)
            cache.set(CACHE_REQUEST_HITS, 0)

    if COUNT_UPDATE_CACHE:
        # Count in UpdateCacheMiddleware is enabled
        LOCAL_CACHE_INFO["responses"] = 0
        LOCAL_CACHE_INFO["response hits"] = 0
        if COUNT_IN_CACHE:
            # Count in cache is enabled
            cache.set(CACHE_RESPONSES, 0)
            cache.set(CACHE_RESPONSE_HITS, 0)


init_cache_counting()


def build_cache_key(url, language_code, site_id):
    cache_key = f"{url}:{language_code}:{site_id}"
    if EXTRA_DEBUG:
        logger.debug(f"Cache key: {cache_key!r}")
    return cache_key


def get_cache_key(request):
    """
    Build the cache key based on the url and:

    * LANGUAGE_CODE: The language code in the url can be different than the
        used language for gettext translation.
    * SITE_ID: request.path is the url without the domain name. So the same
        url in site A and B would result in a collision.
    """
    url = request.get_full_path()

    try:
        language_code = request.LANGUAGE_CODE  # set in django.middleware.locale.LocaleMiddleware
    except AttributeError:
        etype, evalue, etb = sys.exc_info()
        evalue = etype(f"{evalue} (django.middleware.locale.LocaleMiddleware must be insert before cache middleware!)")
        raise etype(evalue).with_traceback(etb)

    site_id = settings.SITE_ID
    cache_key = build_cache_key(url, language_code, site_id)
    return cache_key


def delete_cache_item(url, language_code, site_id=None):
    if site_id is None:
        site_id = settings.SITE_ID

    cache_key = build_cache_key(url, language_code, site_id)
    logger.debug(f"delete from cache: {cache_key!r}")
    cache.delete(cache_key)


class CacheMiddlewareBase:
    def use_cache(self, request, response=None):
        if request.method not in ('GET', 'HEAD'):
            logger.debug(f"Don't cache {request.method!r} ({request.get_full_path()})")
            return False

        if RUN_WITH_DEV_SERVER and request.path.startswith(settings.STATIC_URL):
            if EXTRA_DEBUG:
                logger.debug("Don't cache static files in dev server")
            return False

        if response and response.status_code != 200:
            logger.debug(f"Don't cache response with status code: {response.status_code} ({request.get_full_path()})")
            return False

        if CACHE_MIDDLEWARE_ANONYMOUS_ONLY and request.user.is_authenticated:
            logger.debug("Don't cache requests from authenticated users.")
            return False

        if hasattr(request, '_messages') and len(request._messages) != 0:
            msg = "Don't cache: page for anonymous users has messages."
            if settings.DEBUG:
                storage = messages.get_messages(request)
                raw_messages = ", ".join([message.message for message in storage])
                storage.used = False
                msg += f" (messages: {raw_messages})"
            logger.debug(msg)
            return False

        if response and getattr(response, 'csrf_processing_done', False):
            logger.debug(
                "Don't cache because response.csrf_processing_done==True"
                " (e.g.: view use @csrf_protect decorator)"
            )
            return False

        if cache_callback:
            return cache_callback(request, response)

        return True


def save_incr(key, default=1):
    try:
        cache.incr(key)
    except ValueError:  # Doesn't exist, yet.
        cache.set(key, default)


class FetchFromCacheMiddleware(CacheMiddlewareBase):
    def _count_requests(self, request):
        if RUN_WITH_DEV_SERVER and request.path.startswith(settings.STATIC_URL):
            return

        LOCAL_CACHE_INFO["requests"] += 1
        if COUNT_IN_CACHE:
            save_incr(CACHE_REQUESTS)

    def _count_hit(self):
        LOCAL_CACHE_INFO["request hits"] += 1
        if COUNT_IN_CACHE:
            save_incr(CACHE_REQUEST_HITS)

    def process_request(self, request):
        """
        Try to fetch response from cache, if exists.
        """
        if COUNT_FETCH_FROM_CACHE:
            self._count_requests(request)

        if not self.use_cache(request):
            if EXTRA_DEBUG:
                logger.debug(f"Don't fetch from cache: {request.get_full_path()}")
            return

        cache_key = get_cache_key(request)
        response = cache.get(cache_key)
        if response is None:
            logger.debug(f"Not found in cache: {cache_key!r}")
        else:
            logger.debug(f"Use {cache_key!r} from cache!")
            if COUNT_FETCH_FROM_CACHE:
                self._count_hit()
            response._from_cache = True
            return response


class UpdateCacheMiddleware(CacheMiddlewareBase):
    def _count_response(self, request):
        if RUN_WITH_DEV_SERVER and request.path.startswith(settings.STATIC_URL):
            return

        LOCAL_CACHE_INFO["responses"] += 1
        if COUNT_IN_CACHE:
            save_incr(CACHE_RESPONSES)

    def _count_hit(self):
        LOCAL_CACHE_INFO["response hits"] += 1
        save_incr(CACHE_RESPONSE_HITS)

    def process_response(self, request, response):
        if COUNT_UPDATE_CACHE:
            self._count_response(request)

        if getattr(response, "_from_cache", False):
            if COUNT_UPDATE_CACHE:
                self._count_hit()
            logger.debug("response comes from the cache, no need to update the cache")
            return response
        else:
            # used e.g. in unittests
            response._from_cache = False

        if not self.use_cache(request, response):
            if EXTRA_DEBUG:
                logger.debug(f"Don't put to cache: {request.get_full_path()}")
            return response

        # get the timeout from the "max-age" section of the "Cache-Control" header
        timeout = get_max_age(response)
        if timeout is None:
            # use default cache_timeout
            timeout = settings.CACHE_MIDDLEWARE_SECONDS
        elif timeout == 0:
            logger.debug("Don't cache this page (timeout == 0)")
            return response

        # Create a new HttpResponse for the cache, so we can skip existing
        # cookies and attributes like response.csrf_processing_done
        response2 = HttpResponse(
            content=response._container,
            status=200,
            content_type=response['Content-Type'],
        )
        if response.has_header("Content-Language"):
            response2['Content-Language'] = response['Content-Language']

        if settings.DEBUG or RUN_WITH_DEV_SERVER or request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS:
            # Check if we store a {% csrf_token %} into the cache, this should never happen!
            for content in response._container:
                if "csrfmiddlewaretoken" in content:
                    raise AssertionError(f"csrf_token would be put into the cache! content: {content!r}")

        # Adds ETag, Last-Modified, Expires and Cache-Control headers
        patch_response_headers(response2, timeout)

        cache_key = get_cache_key(request)
        cache.set(cache_key, response2, timeout)

        logger.debug(f"Put to cache: {cache_key!r}")
        return response



"""
    :created: 2018 by Jens Diemer
    :copyleft: 2018-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from email.mime.image import MIMEImage

from django.utils.text import Truncator


def print_mailbox(outbox, max_length=120):
    """
    Usefull for debugging tests, e.g.:

        # ... do something that send mails
        print_mailbox(mail.outbox)
        self.assertEqual(len(mail.outbox), 2)
    """

    def cutted_bytes(data, num=80):
        return Truncator(repr(data)).chars(num=num)

    print("There are %i mails" % len(outbox))
    for no, mail_instance in enumerate(outbox, start=1):
        print("_" * 79)
        print(f" *** Mail No. {no:d}: ***")

        attr_names = [
            attr_name for attr_name in dir(mail_instance) if not attr_name.startswith("_") and attr_name != "body"
        ]
        attr_names.sort()
        attr_names.append("body")

        for attr_name in attr_names:
            attr = getattr(mail_instance, attr_name, "--")

            if callable(attr) or not isinstance(attr, (str, list, tuple, bool)):
                continue

            print("%20s:" % attr_name, end=" ", flush=True)

            if attr_name == "alternatives":
                for data in attr:
                    print(cutted_bytes(data, num=max_length))

            elif attr_name == "attachments":
                print("\n")
                for attachment in attr:
                    if isinstance(attachment, MIMEImage):
                        line = f" * {attachment!r}"
                        data = attachment.get_payload(decode=True)
                    else:
                        filename, data, mine_type = attachment
                        line = f" * {filename} ({mine_type})"

                    print(line, end=" ")
                    print(cutted_bytes(data, num=(max_length - len(line))))
                print()

            elif attr_name == "body":
                print(attr)
                print("-" * max_length)

            else:
                if len(attr) > max_length:
                    attr = cutted_bytes(data, num=max_length)
                print(attr)

        print("=" * 120)



"""
    :created: 22.03.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging

from django.conf import settings
from django.utils.text import slugify

# https://github.com/jedie/django-tools
from django_tools.fixture_tools.languages import iter_languages


log = logging.getLogger(__name__)


class ParlerDummyGenerator:
    """
    Helper to generate dummies for Parler models.
    """
    languages = settings.LANGUAGES  # Languages for created content.
    default_language_code = settings.LANGUAGE_CODE  # First language to start create

    def __init__(self, ParlerModelClass, publisher_model=False, unique_translation_field="slug"):
        self.ParlerModelClass = ParlerModelClass
        self.class_name = ParlerModelClass.__name__
        self.publisher_model = publisher_model  # ParlerModelClass is sub class from PublisherModelBase) ?
        self.unique_translation_field = unique_translation_field

    def get_unique_translation_field_value(self, no, language_code):
        """
        Intended to be overwritten.
        """
        return slugify(
            f"{self.class_name} {language_code} {no:d}"
        )

    def add_instance_values(self, instance, language_code, lang_name, no):
        """
        Intended to be overwritten.
        """
        return instance

    def _get_lookup_kwargs(self, language_code, translation_field_value):
        lookup_kwargs = {
            "translations__language_code": language_code,
            f"translations__{self.unique_translation_field}": translation_field_value
        }
        if self.publisher_model:
            # Is a Publisher model: filter drafts only:
            lookup_kwargs["publisher_is_draft"] = True

        return lookup_kwargs

    def get_or_create(self, count):
        for no in range(1, count + 1):
            self._get_or_create(no=no)

    def _get_or_create(self, no):
        translation_field_value = self.get_unique_translation_field_value(no, language_code=self.default_language_code)

        lookup_kwargs = self._get_lookup_kwargs(
            language_code=self.default_language_code,
            translation_field_value=translation_field_value
        )

        qs = self.ParlerModelClass.objects.language(language_code=self.default_language_code)
        instance, created = qs.get_or_create(**lookup_kwargs)
        if created:
            log.info(" *** New %r (no:%i) created.", self.class_name, no)
        else:
            log.info(" *** Use existing %r (no:%i) instance.", self.class_name, no)

        for language_code, lang_name in iter_languages(languages=None):
            log.info("Add %r instance values", language_code)
            instance.set_current_language(language_code)
            translation_field_value = self.get_unique_translation_field_value(no, language_code)
            setattr(instance, self.unique_translation_field, translation_field_value)
            self.add_instance_values(instance, language_code, lang_name, no)

        log.info("Save no.%i: %r", no, instance)
        instance.save()
        if self.publisher_model:
            log.info("Publish no.%i: %r", no, instance)
            instance.publish()



"""
    print SQL
    ~~~~~~~~~

    :copyleft: 2012-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.db import DEFAULT_DB_ALIAS, connections
from django.test.utils import CaptureQueriesContext
from django.utils.encoding import smart_text


PFORMAT_SQL_KEYWORDS = ("FROM", "WHERE", "ORDER BY", "VALUES")


def pformat_sql(sql):
    # remove unicode u''
    sql = sql.replace("u'", "'").replace('u"', '"')

    sql = sql.replace('`', '')
    for keyword in PFORMAT_SQL_KEYWORDS:
        sql = sql.replace(f' {keyword} ', f'\n\t{keyword} ')

    return smart_text(sql)


class PrintQueries(CaptureQueriesContext):
    """
    with context manager to print the used SQL Queries.
    usage e.g.:
    ---------------------------------------------------------------------------

    from django_tools.unittest_utils.print_sql import PrintQueries

    # e.g. use in unittests:
    class MyTests(TestCase):
        def test_foobar(self):
            with PrintQueries("Create object"):
                FooBar.objects.create("name"=foo)

    # e.g. use in views:
    def my_view(request):
        with PrintQueries("Create object"):
            FooBar.objects.create("name"=foo)

    the output is like:
    ___________________________________________________________________________
     *** Create object ***
    1 - INSERT INTO "foobar" ("name")
        VALUES (foo)
    ---------------------------------------------------------------------------
    """

    def __init__(self, headline, **kwargs):
        self.headline = headline

        if "connection" not in kwargs:
            using = kwargs.pop("using", DEFAULT_DB_ALIAS)
            kwargs["connection"] = connections[using]

        super().__init__(**kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is not None:
            return

        print()
        print("_" * 79)
        if self.headline:
            print(f" *** {self.headline} ***")
        for no, q in enumerate(self.captured_queries, 1):
            sql = pformat_sql(q["sql"])
            msg = smart_text(f"{no:d} - {sql}\n")
            print(msg)
        print("-" * 79)



"""
    Show responses in webbrowser
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Helper functions for displaying test responses in webbrowser.

    :copyleft: 2009-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import cgi
import datetime
import logging
import re
import sys
import tempfile
import traceback
import webbrowser
from collections import OrderedDict
from pprint import pformat
from xml.sax.saxutils import escape

from django.contrib import messages
from django.utils.encoding import force_str
from django.utils.html import strip_tags


try:
    from django.views.debug import get_default_exception_reporter_filter

    get_safe_settings = get_default_exception_reporter_filter().get_safe_settings
except ImportError:
    # Old Django version
    from django.views.debug import get_safe_settings

from django_tools.utils.stack_info import get_stack_info


log = logging.getLogger(__name__)


# Variable to save if the browser was opend in the past.
BROWSER_TRACEBACK_OPENED = False

RESPONSE_INFO_ATTR = ("request", "cookies", "status_code", "_headers", "context", "content")

TEMP_NAME_PREFIX = "django_tools_browserdebug_"
TEMP_DATETIME_FORMAT = "%Y%m%d-%H%M%S_"


def filter_html(content):
    content = re.sub(r"(.*?<body>)", "", content)  # strip head
    content = re.sub(r"(</body>.*?)", "", content)  # strip footer

    # remove style/script blocks:
    content = re.sub(r"(<style.*?>.*?</style>)", "", content, flags=re.IGNORECASE | re.DOTALL)
    content = re.sub(r"(<script.*?>.*?</script>)", "", content, flags=re.IGNORECASE | re.DOTALL)

    content = strip_tags(content)

    # Strip empty lines:
    content = "\n".join([line.rstrip() for line in content.splitlines() if line.strip()])

    return content


def debug_response(response, browser_traceback=True, msg="", display_tb=True, dir=None, print_filtered_html=False):
    """
    Display the response content with a error traceback in a webbrowser.
    TODO: We should delete the temp files after viewing!
    """
    global BROWSER_TRACEBACK_OPENED
    if browser_traceback is not True or BROWSER_TRACEBACK_OPENED:
        return
    # Save for the next traceback
    BROWSER_TRACEBACK_OPENED = True

    content = response.content.decode("utf-8")

    if print_filtered_html:
        print("=" * 79)
        print(filter_html(content))
        print("=" * 79)

    url = response.request["PATH_INFO"]

    stack_info = get_stack_info(filepath_filter="django_tools")
    stack_info.append(msg)
    if display_tb:
        print("\ndebug_response:")
        print("-" * 80)
        print("\n".join(stack_info))
        print("-" * 80)

    stack_info = escape("".join(stack_info))

    response_info = "<dl>\n"

    # -------------------------------------------------------------------------

    response_info += "\t<dt><h3>template</h3> (Without duplicate entries)</dt>\n"
    if hasattr(response, "templates") and response.templates:
        try:
            templates = response.templates
        except AttributeError:
            templates = "---"
        else:
            templates = [template.name for template in OrderedDict.fromkeys(templates)]
            templates = pformat(templates)
            if print_filtered_html:
                print(f"Used template: {response.templates[0].name}")
    else:
        templates = "---"
    response_info += f"\t<dd><pre>{templates}</pre></dd>\n"

    # -------------------------------------------------------------------------

    response_info += "\t<dt><h3>messages</h3></dt>\n"
    msg = messages.get_messages(response.request)
    if msg:
        msg = "".join(["%s\n" % x for x in msg])
    else:
        msg = "---"
    response_info += f"\t<dd><pre>{msg}</pre></dd>\n"

    # -------------------------------------------------------------------------

    # FIXME: Is there a easier way to collect POST data?!?

    response_info += "\t<dt><h3>request.POST</h3></dt>\n"
    try:
        pake_payload = response.request["wsgi.input"]  # django.test.client.FakePayload instance
        payload = pake_payload._FakePayload__content
        payload.seek(0)

        pdict = {"boundary": b"BoUnDaRyStRiNg"}
        post_data = cgi.parse_multipart(payload, pdict)

        for k, v in post_data.items():
            post_data[k] = [v.decode("UTF-8") for v in v]

        post_data = pformat(post_data)
    except Exception as err:
        log.error("Can't collect POST data: %s", err)
        post_data = f"(Error: {err})"

    response_info += f"\t<dd><pre>{post_data}</pre></dd>\n"

    # -------------------------------------------------------------------------

    for attr in RESPONSE_INFO_ATTR:
        # FIXME: There must be exist a easier way to display the info
        response_info += f"\t<dt><h3>{attr}</h3></dt>\n"
        value = getattr(response, attr, "---")
        value = pformat(value)

        try:
            value = force_str(value, errors="strict")
        except UnicodeDecodeError:
            log.exception(f"decode error in attr {attr!r}:")
            value = force_str(value, errors="replace")

        value = escape(value)
        response_info += f"\t<dd><pre>{value}</pre></dd>\n"

    # -------------------------------------------------------------------------

    response_info += "\t<dt><h3>settings</h3></dt>\n"

    try:
        safe_settings = pformat(get_safe_settings())
    except Exception:
        # e.g.: https://github.com/andymccurdy/redis-py/issues/995
        safe_settings = traceback.format_exception(*sys.exc_info())

    response_info += f"\t<dd><pre>{safe_settings}</pre></dd>\n"

    response_info += "</dl>\n"

    # -------------------------------------------------------------------------

    if "</body>" in content:
        info = (
            "\n<br /><hr />\n"
            "<h2>Unittest info</h2>\n"
            "<dl>\n"
            "<dt><h3>url:</h3></dt><dd>%s</dd>\n"
            "<dt><h3>traceback:</h3></dt><dd><pre>%s</pre></dd>\n"
            "<dt><h2>response info:</h2></dt>%s\n"
            "</dl>\n"
            "</body>"
        ) % (url, stack_info, response_info)
        content = content.replace("</body>", info)
    else:
        # Not a html page?
        content += "\n<pre>\n"
        content += "-" * 79
        content += (
            "\nUnittest info\n" "=============\n" "url: %s\n" "traceback:\n%s\n</pre>" "response info:\n%s\n"
        ) % (url, stack_info, response_info)

    prefix = TEMP_NAME_PREFIX
    dt = datetime.datetime.now()
    prefix += dt.strftime(TEMP_DATETIME_FORMAT)
    with tempfile.NamedTemporaryFile(dir=dir, prefix=prefix, suffix=".html", delete=False) as f:
        f.write(content.encode("utf-8"))

        print(f"\nWrite time file '{f.name}'")

        url = f"file://{f.name}"
        print(f"\nDEBUG html page in Browser! (url: {url})")
        try:
            webbrowser.open(url)
        except Exception as err:
            log.error("Can't open browser: %s", err)

        return f.name



"""
    database_info manage command


    setup:

        INSTALLED_APPS = [
            ...
            'django_tools',
            ...
        ]


    usage:

        $ ./manage.py database_info


    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :created: 2017 by Jens Diemer
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from pprint import pprint

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections


class Command(BaseCommand):
    help = "Information about the used database and connections"

    def handle(self, *args, **options):
        self.stdout.write("")
        self.stdout.write("_" * 79)
        self.stdout.write(self.help)
        self.stdout.write("")

        # self.stdout.write("settings.DATABASES = ", ending="")
        # pprint(settings.DATABASES)
        # self.stdout.write("-"*79)

        for alias, settings_dict in settings.DATABASES.items():
            self.stdout.write(f"Database alias {alias!r}:")
            self.stdout.write("")
            engine = settings_dict["ENGINE"]
            engine = engine.rsplit(".", 1)[-1]
            self.stdout.write(f"engine...............: {engine!r}")
            if engine == "sqlite3":
                # https://docs.python.org/3/library/sqlite3.html#module-functions-and-constants
                import sqlite3
                self.stdout.write(f"sqlite lib version...: {sqlite3.sqlite_version!r}")
                self.stdout.write(f"sqlite module version: {sqlite3.version!r}")

            self.stdout.write(f"name.................: {settings_dict['NAME']!r}")
            self.stdout.write(f"user.................: {settings_dict['USER']!r}")
            self.stdout.write(f"host.................: {settings_dict['HOST']!r}")
            self.stdout.write(f"port.................: {settings_dict['PORT']!r}")

        connection_list = connections.all()

        self.stdout.write("")
        self.stdout.write("There are %i connections." % len(connection_list))

        for no, conn in enumerate(connection_list, 1):
            self.stdout.write("")
            self.stdout.write(f"connection {no:d} alias: {conn.alias!r} settings_dict:")
            pprint(conn.settings_dict)

        self.stdout.write("")
        self.stdout.write("-" * 79)



class DisableMigrations:
    """
    Speedup test run start by disable migrations, just create the database tables ;)

    Usage in test settings.py:

    from django_tools.unittest_utils.disable_migrations import DisableMigrations
    MIGRATION_MODULES = DisableMigrations()
    """

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return "notmigrations"



import logging

import django.dispatch
from django.conf import settings
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.views.generic.base import View
from django.views.static import serve

from django_tools.serve_media_app.models import UserMediaTokenModel


logger = logging.getLogger(__name__)

serve_file_request = django.dispatch.Signal()


class UserMediaView(View):
    """
    Serve MEDIA_URL files, but check the current user:
    """

    def check_access_permission(self, request, user_token):
        if request.user.id is None:
            logger.error('Anonymous try to access user media files')
            raise PermissionDenied

        user = request.user

        token = UserMediaTokenModel.objects.get_user_token(user=user)
        if token is None:
            logger.error('Current user %s has no token!', user)
            raise SuspiciousOperation('No user token!')

        if token != user_token:
            # A user tries to access a file from a other user?
            logger.error(f'Wrong user (%s) token: %r is not %r', user, token, user_token)
            raise PermissionDenied

    def get(self, request, user_token, path):
        media_path = f'{user_token}/{path}'
        logger.debug('Serve: %r', media_path)

        if not request.user.is_superuser:
            self.check_access_permission(request, user_token)

        serve_file_request.send(
            sender=self.__class__,
            user=request.user,
            path=path,
            media_path=media_path
        )

        # Send the file to the user:
        return serve(
            request,
            path=media_path,
            document_root=settings.MEDIA_ROOT,
            show_indexes=False
        )



"""
    debug middlewares
    ~~~~~~~~~~~~~~~~~

    more information in the README.

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.conf import settings


class SetRequestDebugMiddleware:
    """
    add 'debug' bool attribute to request object

    debug is on if:
        settings.DEBUG == True
    *OR*
        remote IP is in settings.INTERNAL_IPS
    """

    def process_request(self, request):
        request.debug = settings.DEBUG or request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS



import logging


log = logging.getLogger(__name__)


class LogRequestHeadersMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        for key, value in request.environ.items():
            if key.startswith('HTTP_'):
                log.info(f'{key[5:].lower():<25}: {value!r}')

        return self.get_response(request)



"""
    some additional template filters
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2007-2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import datetime

from django.template.defaultfilters import stringfilter
from django.utils.translation import gettext as _


CHMOD_TRANS_DATA = (
    "---", "--x", "-w-", "-wx", "r--", "r-x", "rw-", "rwx"
)


def chmod_symbol(octal_value):
    """
    Transform a os.stat().st_octal_value octal value to a symbolic string.
    ignores meta information like SUID, SGID or the Sticky-Bit.
    e.g. 40755 -> rwxr-xr-x
    >>> chmod_symbol(644)
    'rw-r--r--'
    >>> chmod_symbol(40755)
    'rwxr-xr-x'
    >>> chmod_symbol("777")
    'rwxrwxrwx'
    """
    octal_value_string = str(octal_value)[-3:]  # strip "meta info"
    return ''.join(CHMOD_TRANS_DATA[int(num)] for num in octal_value_string)


chmod_symbol.is_safe = True
chmod_symbol = stringfilter(chmod_symbol)


def get_oct(value):
    """
    Convert an integer number to an octal string.
    """
    try:
        return oct(value)
    except BaseException:
        return value


get_oct.is_safe = False


def human_duration(t):
    """
    Converts a time duration into a friendly text representation.

    >>> human_duration("type error")
    Traceback (most recent call last):
        ...
    TypeError: human_duration() argument must be timedelta, integer or float)


    >>> human_duration(datetime.timedelta(microseconds=1000))
    '1.0 ms'
    >>> human_duration(0.01)
    '10.0 ms'
    >>> human_duration(0.9)
    '900.0 ms'
    >>> human_duration(datetime.timedelta(seconds=1))
    '1.0 sec'
    >>> human_duration(65.5)
    '1.1 min'
    >>> human_duration(59 * 60)
    '59.0 min'
    >>> human_duration(60*60)
    '1.0 hours'
    >>> human_duration(1.05*60*60)
    '1.1 hours'
    >>> human_duration(datetime.timedelta(hours=24))
    '1.0 days'
    >>> human_duration(2.54 * 60 * 60 * 24 * 365)
    '2.5 years'
    """
    if isinstance(t, datetime.timedelta):
        t = t.total_seconds()
    elif not isinstance(t, (int, float)):
        raise TypeError("human_duration() argument must be timedelta, integer or float)")

    chunks = (
        (60 * 60 * 24 * 365, _('years')),
        (60 * 60 * 24 * 30, _('months')),
        (60 * 60 * 24 * 7, _('weeks')),
        (60 * 60 * 24, _('days')),
        (60 * 60, _('hours')),
    )

    if t < 1:
        return _("%.1f ms") % round(t * 1000, 1)
    if t < 60:
        return _("%.1f sec") % round(t, 1)
    if t < 60 * 60:
        return _("%.1f min") % round(t / 60, 1)

    for seconds, name in chunks:
        count = t / seconds
        if count >= 1:
            count = round(count, 1)
            break
    return f"{count:.1f} {name}"


human_duration.is_safe = True



"""
    $ ./manage.py list_models

"""


from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Just list all existing models in app_label.ModelName format.'

    def handle(self, *args, **options):
        self.stdout.write("\nexisting models in app_label.ModelName format:\n\n")

        dotnames = []
        app_configs = apps.get_app_configs()
        for app_config in app_configs:
            app_label = app_config.label
            models = app_config.get_models()
            for model in models:
                dotnames.append(
                    f"{app_label}.{model._meta.object_name}"
                )

        for no, dotname in enumerate(sorted(dotnames), 1):
            self.stdout.write(f"{no:02d} - {dotname}\n")

        self.stdout.write("\nINSTALLED_APPS....: %i\n" % len(settings.INSTALLED_APPS))
        self.stdout.write("Apps with models..: %i\n\n" % len(app_configs))



"""
    django-tools import helper
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    additional helper to the existing django.utils.importlib

    :copyleft: 2012-2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import logging
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


logger = logging.getLogger("DjangoToolsImportLib")


def get_attr_from_string(path, obj_name=""):
    """
    Return a attribute from a module by the given path string.

    >>> get_attr_from_string("django_tools.cache.site_cache_middleware.UpdateCacheMiddleware")
    <class 'django_tools.cache.site_cache_middleware.UpdateCacheMiddleware'>
    """
    try:
        module_name, class_name = path.rsplit('.', 1)
    except ValueError:
        raise ImproperlyConfigured(f"{path} isn't a {obj_name} module")

    try:
        mod = import_module(module_name)
    except ImportError as e:
        raise ImproperlyConfigured(f'Error importing {obj_name} module {module_name}: "{e}"')

    try:
        attr = getattr(mod, class_name)
    except AttributeError:
        raise ImproperlyConfigured(f'{obj_name} module "{module_name}" does not define a "{class_name}" class')

    return attr


def get_class_instance(path, obj_name=""):
    """
    Returns a class instance from a module by the given path string.

    >>> get_class_instance("django_tools.cache.site_cache_middleware.UpdateCacheMiddleware") # doctest: +ELLIPSIS
    <django_tools.cache.site_cache_middleware.UpdateCacheMiddleware object at ...>
    """
    class_obj = get_attr_from_string(path, obj_name)
    class_instance = class_obj()
    return class_instance


def get_setting(setting_name):
    """
    return the settings value, create debug log if not set/empty.

    >>> get_setting("EMAIL_BACKEND")
    'django.core.mail.backends.locmem.EmailBackend'
    """
    path = getattr(settings, setting_name, None)
    if path:
        return path
    else:
        if not hasattr(settings, setting_name):
            logger.debug(f"{setting_name!r} not in settings defined")
        else:
            logger.debug(f"settings.{setting_name} is None or empty")


def get_attr_from_settings(setting_name, obj_name=""):
    """
    returns a attribute from the given settings path string.

    >>> get_attr_from_settings("EMAIL_BACKEND", "email backend")
    <class 'django.core.mail.backends.locmem.EmailBackend'>
    """
    path = get_setting(setting_name)
    if path:
        return get_attr_from_string(path, obj_name)


def get_class_instance_from_settings(setting_name, obj_name=""):
    """
    returns a class instance from the given settings path string.

    >>> get_class_instance_from_settings("EMAIL_BACKEND", "email backend") # doctest: +ELLIPSIS
    <django.core.mail.backends.locmem.EmailBackend object at ...>
    """
    path = get_setting(setting_name)
    if path:
        return get_class_instance(path, obj_name)



"""
    sign separated
    ~~~~~~~~~~~~~~

     * model field
     * form field
     * widget

    :copyleft: 2010-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django import forms
from django.db import models


def _split(raw_value, separator, strip_items, skip_empty):
    if not raw_value:
        return ()

    if isinstance(raw_value, (list, tuple)):
        return raw_value

    values = []
    for item in raw_value.split(separator):
        if strip_items:
            item = item.strip()

        if item in values or skip_empty and not item:
            continue

        values.append(item)

    values = tuple(values)
    return values


def _join(value, separator):
    if value is None:
        value = ""
    elif isinstance(value, (list, tuple)):
        value = separator.join(value)
    return value


class SignSeparatedInput(forms.widgets.Input):
    input_type = 'text'

    def __init__(self, separator=",", *args, **kwargs):
        self.separator = separator
        super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        value = _join(value, self.separator)
        return super().render(name, value, attrs)


class SignSeparatedFormField(forms.CharField):
    """
    >>> SignSeparatedFormField().clean("one, two")
    ('one', 'two')

    >>> SignSeparatedFormField().to_python("one , two, 3,4")
    ('one', 'two', '3', '4')
    >>> SignSeparatedFormField(strip_items=False).clean("one , two, 3,4")
    ('one ', ' two', ' 3', '4')

    >>> SignSeparatedFormField(separator=" ").clean("one  two 3")
    ('one', 'two', '3')
    >>> SignSeparatedFormField(separator=" ", skip_empty=False).clean("one  two 3")
    ('one', '', 'two', '3')

    >>> try:
    ...     SignSeparatedFormField().clean(None)
    ... except Exception as err:
    ...     print(err.__class__.__name__, err)
    ValidationError ['This field is required.']

    >>> try:
    ...     SignSeparatedFormField().clean("")
    ... except Exception as err:
    ...     print(err.__class__.__name__, err)
    ValidationError ['This field is required.']
    """

    def __init__(self, separator=",", strip_items=True, skip_empty=True, *args, **kwargs):
        self.separator = separator
        self.strip_items = strip_items
        self.skip_empty = skip_empty

        self.widget = SignSeparatedInput(separator)

        super().__init__(*args, **kwargs)

    def to_python(self, value):
        values = _split(value, self.separator, self.strip_items, self.skip_empty)
        return values


# @six.add_metaclass(models.SubfieldBase)
class SignSeparatedModelField(models.TextField):
    """
    A dict field.
    Stores a python dict into a text field.

    >>> SignSeparatedModelField().to_python("foo, bar")
    ('foo', 'bar')

    >>> SignSeparatedModelField().get_db_prep_save(('foo', 'bar'))
    'foo,bar'

    >>> f = SignSeparatedModelField().formfield()
    >>> isinstance(f, SignSeparatedFormField)
    True
    >>> f.clean("one , two, 3,4")
    ('one', 'two', '3', '4')

    kwargs would be pass to the widget:
    >>> f = SignSeparatedModelField(separator="x", strip_items=False, skip_empty=False).formfield()
    >>> f.clean("1x2x x 3")
    ('1', '2', ' ', ' 3')


    >>> from django.db import models
    >>> from django.forms.models import ModelForm

    >>> class TestModel(models.Model):
    ...     test = SignSeparatedModelField(separator=";")
    ...     class Meta:
    ...         default_permissions = () # Don't create permissions, see below
    ...         app_label = "django_tools"

    >>> class TestForm(ModelForm):
    ...     class Meta:
    ...         model = TestModel
    ...         exclude=()

    >>> f = TestForm({'test': None})
    >>> f.is_valid()
    False
    >>> f = TestForm({'test': ""})
    >>> f.is_valid()
    False
    >>> f = TestForm({'test': "one; two;three"})
    >>> f.is_valid()
    True
    >>> f.cleaned_data
    {'test': ('one', 'two', 'three')}

    Note: TestModel.Meta.default_permissions is only set, because otherwise other unittests will failed!
    Django will automatically create new entries in auth.models.Permission and this is not
    atomic in doc test run :(
    """

    def __init__(self, separator=",", strip_items=True, skip_empty=True, *args, **kwargs):
        self.separator = separator
        self.strip_items = strip_items
        self.skip_empty = skip_empty
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        """
        Converts the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.
        """
        values = _split(value, self.separator, self.strip_items, self.skip_empty)
        return values

    def get_db_prep_save(self, value, **kwargs):
        "Returns field's value prepared for saving into a database."
        value = _join(value, self.separator)
        return value

    def formfield(self, **kwargs):
        """ Use always own widget and form field. """
        kwargs["separator"] = self.separator
        kwargs["strip_items"] = self.strip_items
        kwargs["skip_empty"] = self.skip_empty

        kwargs["widget"] = SignSeparatedInput
        kwargs["form_class"] = SignSeparatedFormField
        return super().formfield(**kwargs)



import io
from unittest import mock

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.core.files.storage import DefaultStorage
from django.http import FileResponse
from django.test import TestCase
from django.urls import reverse
from model_bakery import baker
from override_storage import locmem_stats_override_storage

from django_tools.serve_media_app.constants import USER_TOKEN_LENGTH
from django_tools.serve_media_app.models import UserMediaTokenModel, generate_media_path
from django_tools.serve_media_app.views.serve_user_files import serve_file_request
from django_tools.unittest_utils.signals import SignalsContextManager


class SignalsTestCase(TestCase):
    def test_create_user_token_via_signal(self):
        assert UserMediaTokenModel.objects.count() == 0
        UserModel = get_user_model()
        user = baker.make(UserModel)
        assert UserMediaTokenModel.objects.count() == 1

        instance = UserMediaTokenModel.objects.get_from_user(user)
        assert instance is not None
        token1 = instance.token
        assert len(token1) == USER_TOKEN_LENGTH

        user.save()  # Token must be not changed
        user = UserModel.objects.get(pk=user.pk)
        instance = UserMediaTokenModel.objects.get_from_user(user)
        assert instance.token == token1

    @locmem_stats_override_storage()
    def test_serve_file_request_signal(self):
        with mock.patch('secrets.token_urlsafe', return_value='ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
            super_user = baker.make(get_user_model(), is_superuser=True, email='super@user.tld')
        self.client.force_login(super_user)

        with mock.patch('secrets.token_urlsafe', return_value='ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
            media_path = generate_media_path(user=super_user, filename='foobar.ext')

        assert media_path == 'abcdefghijkl/abcdefghijklmnopqrstuvwx/foobar.ext'
        storage = DefaultStorage()
        content = io.BytesIO('Test äöüß !'.encode())
        storage.save(media_path, content)

        assert settings.MEDIA_URL == '/media/'
        url = reverse(
            'serve_media_app:serve-media',
            kwargs={'user_token': 'abcdefghijkl', 'path': 'abcdefghijklmnopqrstuvwx/foobar.ext'}
        )
        assert url == '/media/abcdefghijkl/abcdefghijklmnopqrstuvwx/foobar.ext'

        # Test with a signal callback that will not allow the access:

        def deny_access_callback(user, path, media_path, **kwargs):
            assert user.email == 'super@user.tld'
            assert path == 'abcdefghijklmnopqrstuvwx/foobar.ext'
            assert media_path == 'abcdefghijkl/abcdefghijklmnopqrstuvwx/foobar.ext'
            raise PermissionDenied

        with SignalsContextManager(serve_file_request, deny_access_callback):
            response = self.client.get(url)
            assert response.status_code == 403

        # Test with a signal callback that allowes the access:

        def allow_access_callback(user, path, media_path, **kwargs):
            assert user.email == 'super@user.tld'
            assert path == 'abcdefghijklmnopqrstuvwx/foobar.ext'
            assert media_path == 'abcdefghijkl/abcdefghijklmnopqrstuvwx/foobar.ext'

        with SignalsContextManager(serve_file_request, allow_access_callback):
            response = self.client.get(url)
            assert response.status_code == 200
            assert isinstance(response, FileResponse)
            assert response.getvalue().decode('UTF-8') == 'Test äöüß !'



"""
    :created: 23.12.2018 by Jens Diemer <opensource@jensdiemer.de>
    :copyleft: 2018-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import filecmp
import logging
from pathlib import Path

from django.core.files.storage import FileSystemStorage

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.assertments import assert_is_file


log = logging.getLogger(__name__)


def file_compare(path1, path2):
    log.debug("compare %r vs, %r", path1, path2)
    assert_is_file(path1)
    assert_is_file(path2)

    # check if file refer to same file
    if path1.samefile(path2):
        log.debug("File refer to the same file")
        return True

    # compare file size:
    size1 = path1.stat().st_size
    size2 = path2.stat().st_size
    if size1 != size2:
        log.debug("Not the same file: different file size.")
        return False

    # Compare file content:
    filecmp.clear_cache()  # if we didn't clear the cache: unittests may be failed!
    return filecmp.cmp(str(path1), str(path2), shallow=False)  # str() needed for python 3.5


class OverwriteFileSystemStorage(FileSystemStorage):
    """
    Overwrite existing files. Usage, e.g:

        class ExampleModel(models.Model):
            foo_file = models.FileField(
                storage=OverwriteFileSystemStorage(create_backups=True)
            )
            bar_image = models.ImageField(
                storage=OverwriteFileSystemStorage(create_backups=True)
            )

    Backups:
        * backup made by appending a suffix and sequential number, e.g.:
            * source....: foo.bar
            * backup 1..: foo.bar.bak
            * backup 2..: foo.bar.bak0
            * backup 3..: foo.bar.bak1

    Backup files are only made if file content changed.

    Based on: https://djangosnippets.org/snippets/976/
    see also: https://code.djangoproject.com/ticket/11663
    """

    def __init__(self, *args, create_backups=True, backup_suffix=".bak", **kwargs):
        """
        :param backup: True/False -> create backup before overwrite
        """
        super().__init__(*args, **kwargs)
        self.create_backups = create_backups
        self.backup_suffix = backup_suffix

    def get_available_name(self, name, max_length=None):
        """
        Always return the origin name and never add random characters to get a unique name
        """
        return name

    def get_bak_path(self, full_path):
        for no in range(1, 99):
            bak_path = full_path.with_suffix(f"{full_path.suffix}{self.backup_suffix}{no:02d}")
            if not bak_path.is_file():
                return bak_path

        raise RuntimeError("No free backup filenames!")

    def _save(self, name, content):
        """
        There's a potential race condition here:
        It's possible that two threads might whant to save the same file.
        """
        full_path = Path(self.path(name))

        bak_path = None

        if full_path.is_file():
            if self.create_backups:
                bak_path = self.get_bak_path(full_path)

                log.info("Backup %s -> %s", full_path, bak_path)
                full_path.rename(bak_path)
            else:
                log.info("Remove old file: %s", full_path)
                full_path.unlink()

        # Save content
        name = super()._save(name, content)
        log.warning(repr(content))
        log.debug("Save file as: %r", name)

        if bak_path is not None and file_compare(full_path, bak_path) is True:
            log.info("Remove %r (same file content)", bak_path)
            bak_path.unlink()

        return name



import io
import tempfile
from unittest import mock

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.core.files import File
from django.core.files.storage import DefaultStorage
from django.http import FileResponse
from django.test import Client, TestCase, override_settings
from django.urls import reverse
from model_bakery import baker

from django_tools.serve_media_app.models import UserMediaTokenModel, generate_media_path
from django_tools.unittest_utils.mockup import ImageDummy
from django_tools_test_project.django_tools_test_app.models import UserMediaFiles


class UserMediaViewsTestCase(TestCase):
    def test_basic(self):
        assert settings.MEDIA_URL == '/media/'
        assert get_user_model() == User

        url = reverse('serve_media_app:serve-media', kwargs={'user_token': 'foo', 'path': 'bar'})
        assert url == '/media/foo/bar'

        with tempfile.TemporaryDirectory() as temp:
            with override_settings(MEDIA_ROOT=temp):
                user = baker.make(User, username='owner')
                file_path = generate_media_path(user, filename='foobar.txt')

                storage = DefaultStorage()
                content = io.BytesIO('Test äöüß !'.encode())
                final_file_path = storage.save(file_path, content)
                assert final_file_path == file_path

                url = f'/media/{file_path}'

                # Anonymous has no access:
                response = self.client.get(url)
                assert response.status_code == 403

                # Can't access with wrong user:
                other_user = baker.make(User, username='not-owner')
                self.client.force_login(other_user)
                response = self.client.get(url)
                assert response.status_code == 403

                # Can access with the right user:
                self.client.force_login(user)
                response = self.client.get(url)
                assert response.status_code == 200
                assert isinstance(response, FileResponse)
                assert response.getvalue().decode('UTF-8') == 'Test äöüß !'

                # Test whats happen, if token was deleted
                UserMediaTokenModel.objects.all().delete()
                response = self.client.get(url)
                assert response.status_code == 400  # SuspiciousOperation -> HttpResponseBadRequest

    def test_via_model(self):
        with mock.patch('secrets.token_urlsafe', return_value='ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
            user = baker.make(User, username='owner')

        other_user = baker.make(User, username='not-owner')

        with tempfile.TemporaryDirectory() as temp:
            with override_settings(MEDIA_ROOT=temp):
                with mock.patch('secrets.token_urlsafe', return_value='12345678901234567890'):
                    instance = UserMediaFiles.objects.create(
                        user=user,
                        file=File(io.BytesIO('Test äöüß !'.encode()), name='filename.ext'),
                        image=ImageDummy(width=1, height=1).create_django_file_info_image(text=None),
                    )

                token_instance = UserMediaTokenModel.objects.get(user=user)
                assert repr(token_instance) == (
                    f"<UserMediaTokenModel: user:{user.pk} token:'abcdefghijkl' ({token_instance.pk})>"
                )

                user_token = UserMediaTokenModel.objects.get_user_token(user)
                assert user_token == 'abcdefghijkl'
                other_user_token = UserMediaTokenModel.objects.get_user_token(other_user)
                assert other_user_token != user_token

                urls = (
                    '/media/abcdefghijkl/12345678901234567890/filename.ext',
                    '/media/abcdefghijkl/12345678901234567890/dummy.jpeg'
                )

                assert instance.file.url == urls[0]
                assert instance.image.url == urls[1]

                # Anonymous has no access:
                for url in urls:
                    response = self.client.get(url)
                    assert response.status_code == 403

                # Can't access with wrong user:
                self.client = Client()
                self.client.force_login(other_user)
                for url in urls:
                    response = self.client.get(url)
                    assert response.status_code == 403

                # Can access with the right user:
                self.client = Client()
                self.client.force_login(user)
                response = self.client.get('/media/abcdefghijkl/12345678901234567890/filename.ext')
                assert response.status_code == 200
                assert isinstance(response, FileResponse)
                assert response.getvalue().decode('UTF-8') == 'Test äöüß !'
                response = self.client.get('/media/abcdefghijkl/12345678901234567890/dummy.jpeg')
                assert response.status_code == 200
                assert isinstance(response, FileResponse)
                assert len(response.getvalue())

                # Test whats happen, if token was deleted
                UserMediaTokenModel.objects.all().delete()
                for url in urls:
                    response = self.client.get(url)
                    assert response.status_code == 400  # SuspiciousOperation -> HttpResponseBadRequest



"""
    exceptions for filemanager
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


class FilemanagerError(Exception):
    """
    for errors with a message to staff/admin users.
    e.g.: Gallery filesystem path doesn't exist anymore.
    """
    pass


class DirectoryTraversalAttack(FilemanagerError):
    """
    Some unauthorized signs are found or the path is out of the base path.
    """
    pass



"""
    models utils
    ~~~~~~~~~~~~

    :copyleft: 2009-2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db import IntegrityError
from django.db.models import signals
from django.utils.text import get_text_list
from django.utils.translation import gettext as _


try:
    from django.core.exceptions import FieldDoesNotExist
except ImportError:
    # Old Django version
    from django.db.models.fields import FieldDoesNotExist


def check_unique_together(sender, **kwargs):
    """
    Check models unique_together manually. Because Django will only
    enforced unique together at database level with UNIQUE, but
    some databases (e.g. SQLite) doesn't support this.

    NOTE: SQLite supports UNIQUE since 2.0 (from 2001) !

    usage:
        from django.db.models import signals
        from django_tools.model_utils import check_unique_together
        signals.pre_save.connect(check_unique_together, sender=MyModelClass)

    or use:
        from django_tools.model_utils import auto_add_check_unique_together
        auto_add_check_unique_together(MyModelClass)

    This will add the signal only if a Database doesn't support UNIQUE, see below.
    """
    instance = kwargs["instance"]
    for field_names in sender._meta.unique_together:
        model_kwargs = {}
        for field_name in field_names:
            try:
                data = getattr(instance, field_name)
            except FieldDoesNotExist:
                # e.g.: a missing field, which is however necessary.
                # The real exception on model creation should be raised.
                continue
            model_kwargs[field_name] = data

        query_set = sender.objects.filter(**model_kwargs)
        if instance.pk is not None:
            # Exclude the instance if it was saved in the past
            query_set = query_set.exclude(pk=instance.pk)

        count = query_set.count()
        if count > 0:
            field_names = get_text_list(field_names, _('and'))
            msg = _("%(model_name)s with this %(field_names)s already exists.") % {
                'model_name': str(instance.__class__.__name__),
                'field_names': str(field_names)
            }
            raise IntegrityError(msg)


def auto_add_check_unique_together(model_class):
    """
    Add only the signal handler check_unique_together, if a database without UNIQUE support is used.

    NOTE: SQLite supports UNIQUE since 2.0 (from 2001) !
    """
    try:
        # new setting scheme
        engine = settings.DATABASES["default"]["ENGINE"]
    except AttributeError:
        # fall back to old scheme
        engine = settings.DATABASE_ENGINE

    if "sqlite3" in engine:  # 'postgresql', 'mysql', 'sqlite3' or 'ado_mssql'.
        signals.pre_save.connect(check_unique_together, sender=model_class)



"""
    Local sync cache
    ~~~~~~~~~~~~~~~~

    The Problem:
    If you use a normal dict for cache some values, you can't clear it in
    a multi-threaded environment, because you have no access to the dict in
    other threads.

    The Solution:
    Use LocalSyncCache() as a cache dict. If dict.clear() called, the dict
    in all threads would be cleared.

    How it works:
    * Every cache dict memorize his own creation/reset time.
    * in dict.clear() the reset time would be saved
      into django cache (to share it with all threads)
    * On every request the LocalSyncCacheMiddleware called all existing cache dict
      in the current threads to look into the shared django cache, if they
      are outdatet or not. If they are outdated, the dict would be cleaned.


    usage
    ~~~~~

    Add LocalSyncCacheMiddleware to settings, e.g:
    ---------------------------------------------------------------------------
        MIDDLEWARE_CLASSES = (
            ...
            'django_tools.local_sync_cache.LocalSyncCacheMiddleware.LocalSyncCacheMiddleware',
            ...
        )
    ---------------------------------------------------------------------------


    Create a cache dict with a id.
    Use it in a model, e.g.:
    ---------------------------------------------------------------------------
        from django.db import models
        from django_tools.local_sync_cache.local_sync_cache import LocalSyncCache

        class PageTree(models.Model):
            parent = models.ForeignKey("self", null=True, blank=True)
            slug = models.SlugField()

            _url_cache = LocalSyncCache(id="PageTree_absolute_url") # <<<---
            def get_absolute_url(self):
                if self.pk in self._url_cache:
                    return self._url_cache[self.pk]

                if self.parent:
                    parent_url = self.parent.get_absolute_url()
                    url = parent_url + self.slug + "/"
                else:
                    url = "/" + self.slug + "/"

                self._url_cache[self.pk] = url
                return url

        def save(self, *args, **kwargs):
            self._url_cache.clear() # Clean the local url cache dict
            return super(PageTree, self).save(*args, **kwargs)
    ---------------------------------------------------------------------------


    logging
    ~~~~~~~

    To enable logging, add this to your settings, e.g.:

        import logging
        logger = logging.getLogger("django_tools.local_sync_cache")
        logger.setLevel(log.logging.DEBUG)
        logger.addHandler(log.logging.FileHandler("local_sync_cache.log"))


    :copyleft: 2011-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import datetime
import logging
import sys
import time

from django.conf import settings
from django.core.cache import caches


logger = logging.getLogger("django_tools.local_sync_cache")


LOCAL_SYNC_CACHE_BACKEND = getattr(settings, "LOCAL_SYNC_CACHE_BACKEND", "local_sync_cache")


def _get_cache():
    """
    return django cache object.
    Try to use a own backend and fallback to 'default' if not defined in settings.CACHES
    """
    if LOCAL_SYNC_CACHE_BACKEND not in settings.CACHES:
        # TODO: Not needed in django v1.4: https://code.djangoproject.com/ticket/16410
        msg = f"You should define a '{LOCAL_SYNC_CACHE_BACKEND}' cache in your settings.CACHES (use default cache)"
        logger.critical(msg)
        cache_name = "default"  # fallback to default cache entry
    else:
        cache_name = LOCAL_SYNC_CACHE_BACKEND

    backend = settings.CACHES[cache_name]["BACKEND"]
    if "dummy" in backend or "locmem" in backend:
        msg = f"You should use Memcache, FileBasedCache or DatabaseCache and not: {backend}"
        logger.critical(msg)

    django_cache = caches[cache_name]
    logger.debug(f"Use django '{cache_name}' cache: {django_cache!r}")
    return django_cache


class LocalSyncCache(dict):
    INIT_COUNTER = {}  # Counts how often __init__ used, should always be 1!

    # Stores all existing instance, used in middleware to call check_state()
    CACHES = []

    # Store the last reset times secondary in this local thread.
    _OWN_RESET_TIMES = {}

    def __init__(self, id=None, unique_ids=True):
        if id is None:
            raise AssertionError("LocalSyncCache must take a id as argument.")

        if unique_ids:
            for existing_cache in self.CACHES:
                if id == existing_cache.id:
                    logger.error(
                        "ID {!r} was already used! It must be unique! (Existing ids are: {})".format(
                            id, repr([i.id for i in self.CACHES])
                        )
                    )

        self.id = id
        self.django_cache = _get_cache()
        self.last_reset = time.time()  # Save last creation/reset time
        self.CACHES.append(self)

        if self.id not in self.INIT_COUNTER:
            self.INIT_COUNTER[self.id] = 1
        else:
            logger.error(f"Error: __init__ for {self.id} was called to often!")
            self.INIT_COUNTER[self.id] += 1

        self.request_counter = 0  # Counts how often check_state called (Normally called one time per request)
        self.own_clear_counter = 0  # Counts how often clear called in this thread
        self.ext_clear_counter = 0  # Counts how often clears from external thread

        logger.debug(f"{id!r} __init__")

    def check_state(self):
        """
        Check if we are out-dated or not.
        Should be called at the start of a request. e.g.: by middleware
        """
        self.request_counter += 1
        global_update_time = self.django_cache.get(self.id)

        if global_update_time is None:
            if self.id in self._OWN_RESET_TIMES:
                # clear() was called in the past in this thread and it
                # is not stored in the django cache -> resave it
                logger.info(f"Resave {self.id!r} last reset time in cache")
                self.django_cache.set(self.id, self._OWN_RESET_TIMES[self.id])
        elif self.last_reset < global_update_time:
            # We have out-dated data -> reset dict
            self.ext_clear_counter += 1
            logger.info(
                f"{self.id!r} out-dated data -> reset"
                f" (global_update_time: {global_update_time!r}"
                f" - self.last_reset: {self.last_reset!r})"
            )
            dict.clear(self)
            self.last_reset = time.time()

            if self.id in self._OWN_RESET_TIMES:
                # In this thread clear() was called in the past and now in
                # a other thread clear() was called.
                del(self._OWN_RESET_TIMES[self.id])
                logger.debug(f"remove {self.id!r} from _OWN_RESET_TIMES")

    def clear(self):
        """
        Must be called from the process/thread witch change the data.
            * Clear the dict
            * Save clear time in django cache and in self._OWN_RESET_TIMES
        """
        self.own_clear_counter += 1
        dict.clear(self)
        self.last_reset = time.time()
        self.django_cache.set(self.id, self.last_reset)
        logger.info(f"{self.id!r} - dict.clear - Set global_update_time to {self.last_reset!r}")

        # Save reset time in this thread for re-adding it to cache in check_state()
        self._OWN_RESET_TIMES[self.id] = self.last_reset

        # Check if cache worked
        cached_value = self.django_cache.get(self.id)
        if cached_value != self.last_reset:
            logger.error(f"Cache seems not to work: {cached_value!r} != {self.last_reset!r}")

    @staticmethod
    def get_cache_information():
        cache_information = []
        django_cache = _get_cache()
        for instance in LocalSyncCache.CACHES:
            try:
                instance_size = sys.getsizeof(instance)  # New in version 2.6
            except (AttributeError, TypeError):  # PyPy raised a TypeError
                instance_size = None

            id = instance.id

            cleared = id in LocalSyncCache._OWN_RESET_TIMES
            global_update_time = django_cache.get(id)

            last_reset_datetime = datetime.datetime.fromtimestamp(instance.last_reset)
            if global_update_time:
                global_update_datetime = datetime.datetime.fromtimestamp(global_update_time)
            else:
                global_update_datetime = None

            cache_information.append({
                "instance": instance,
                "length": len(instance),
                "size": instance_size,
                "cleared": cleared,
                "global_update_time": global_update_time,
                "global_update_datetime": global_update_datetime,
                "last_reset_datetime": last_reset_datetime,
                "init_counter": LocalSyncCache.INIT_COUNTER[id],
            })
        return cache_information

    @staticmethod
    def pformat_cache_information():
        output = []
        attributes = ("id", "request_counter", "own_clear_counter", "ext_clear_counter")

        cache_information = LocalSyncCache.get_cache_information()
        for item in cache_information:
            output.append(" -" * 40)

            instance = item["instance"]

            for attr in attributes:
                output.append(f"{attr:>22}: {getattr(instance, attr)}")

            for key, value in item.items():
                output.append(f"{key:>22}: {value!r}")

        return "\n".join(output)



"""
    :copyleft: 2017-2020 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import bleach
from django.utils.html import strip_tags


def html2text(value):
    """
    Returns the text from HTML by strip all HTML tags.
    Used first bleach.clean() and then django 'strip_tags'
    """
    if value:
        value = value.strip()

    if value:
        # use Django
        value = strip_tags(value)

        value = bleach.clean(value)

    return value



from django.urls import path

from django_tools.serve_media_app.views.serve_user_files import UserMediaView


app_name = 'serve_media_app'
urlpatterns = [
    path('<slug:user_token>/<path:path>', UserMediaView.as_view(), name='serve-media'),
]



"""
    utils around time/date/datetime
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2011 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import datetime


def datetime2float(t):
    """
    >>> datetime2float(datetime.timedelta(seconds=1))
    1.0

    >>> f = datetime2float(datetime.timedelta(weeks=2, seconds=20.1, microseconds=100))
    >>> f
    1209620.1001
    >>> f == 20.1 + 0.0001 + 2 * 7 * 24 * 60 * 60
    True

    >>> datetime2float("type error")
    Traceback (most recent call last):
        ...
    TypeError: datetime2float() argument must be a timedelta instance.
    """
    if not isinstance(t, datetime.timedelta):
        raise TypeError("datetime2float() argument must be a timedelta instance.")

    # timedelta.total_seconds() is new in Python 2.7
    return (float(t.microseconds) + (t.seconds + t.days * 24 * 3600) * 1E6) / 1E6



class SignalsContextManager:
    def __init__(self, signal, callback):
        self.signal = signal
        self.callback = callback

    def __enter__(self):
        self.signal.connect(self.callback)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.signal.disconnect(self.callback)



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver

from django_tools.serve_media_app.constants import PATH_TOKEN_LENGTH, USER_TOKEN_LENGTH
from django_tools.serve_media_app.utils import clean_filename, get_random_string


log = logging.getLogger(__name__)


def generate_token():
    return get_random_string(USER_TOKEN_LENGTH)


def generate_media_path(user, filename):
    """
    Generate random MEDIA path for the given user and filename
    """
    user_token = UserMediaTokenModel.objects.get_user_token(user)

    random_string = get_random_string(PATH_TOKEN_LENGTH)
    filename = clean_filename(filename)
    filename = f'{user_token}/{random_string}/{filename}'

    log.info('Upload file path: %r', filename)
    return filename


def user_directory_path(instance, filename):
    """
    FileField/ImageField 'upload_to' handler
    Function for e.g.: models.ImageField(upload_to=user_directory_path,

    Upload to /MEDIA_ROOT/...
    """
    return generate_media_path(user=instance.user, filename=filename)


class UserMediaTokenQuerySet(models.QuerySet):
    def get_from_user(self, instance):
        qs = self.filter(user_id=instance.pk)
        return qs.first()

    def get_user_token(self, user):
        instance = self.get_from_user(instance=user)
        if instance is not None:
            return instance.token


class UserMediaTokenModel(models.Model):
    """
    Store for every User a random token.
    This token will be added to uploaded MEDIA files for
    identify the user.
    Auto created via post_save signal, see below.
    """
    objects = UserMediaTokenQuerySet.as_manager()

    user = models.ForeignKey(  # "Owner" of this entry
        settings.AUTH_USER_MODEL,
        related_name='token',
        on_delete=models.CASCADE,
        editable=False,
    )
    token = models.CharField(
        max_length=USER_TOKEN_LENGTH,
        default=generate_token,
        unique=True,
    )

    def __repr__(self):
        return f'<UserMediaTokenModel: user:{self.user_id} token:{self.token!r} ({self.pk})>'


@receiver(post_save, sender=get_user_model())
def create_token(sender, created=None, instance=None, **kwargs):
    assert instance is not None
    token_instance, created = UserMediaTokenModel.objects.get_or_create(user_id=instance.pk)
    if created:
        log.info('User media token created for: %s', instance)
    else:
        log.debug('User media token exists for: %s', instance)



"""
    :created: 11.12.2018 by Jens Diemer
    :copyleft: 2018-2020 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from django_tools.selenium.chromedriver import SeleniumChromiumTestCase
from django_tools.selenium.geckodriver import SeleniumFirefoxTestCase


class SeleniumChromiumStaticLiveServerTestCase(StaticLiveServerTestCase, SeleniumChromiumTestCase):
    pass


class SeleniumFirefoxStaticLiveServerTestCase(StaticLiveServerTestCase, SeleniumFirefoxTestCase):
    pass



"""
    smooth cache backends
    ~~~~~~~~~~~~~~~~~~~~~

    more information in the README.

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import logging
import os
import time

from django.conf import settings
from django.core.cache.backends.db import DatabaseCache
from django.core.cache.backends.filebased import FileBasedCache
from django.core.cache.backends.locmem import LocMemCache
from django.core.cache.backends.memcached import MemcachedCache, PyLibMCCache


logger = logging.getLogger("django-tools.SmoothCache")
if not logger.handlers:
    # ensures we don't get any 'No handlers could be found...' messages
    logger.addHandler(logging.NullHandler())

# if "runserver" in sys.argv or "tests" in sys.argv:
#    log.logging.basicConfig(format='%(created)f pid:%(process)d %(message)s')
#    logger.setLevel(log.logging.DEBUG)
#    logger.addHandler(log.logging.StreamHandler())


SMOOTH_CACHE_CHANGE_TIME = getattr(settings, "SMOOTH_CACHE_CHANGE_TIME", "DJANGO_TOOLS_SMOOTH_CACHE_CHANGE_TIME")
SMOOTH_CACHE_UPDATE_TIMESTAMP = getattr(settings, "SMOOTH_CACHE_UPDATE_TIMESTAMP", 10)
SMOOTH_CACHE_TIMES = getattr(settings, "SMOOTH_CACHE_TIMES", (
    # load value, max age in sec.
    (0, 5),  # < 0.1 ->  5sec
    (0.1, 10),  # 0.1 - 0.5 -> 10sec
    (0.5, 30),  # 0.5 - 1.0 -> 30sec
    (1.0, 60),  # 1.0 - 1.5 ->  1Min
    (1.5, 120),  # 1.5 - 2.0 ->  2Min
    (2.0, 300),  # 2.0 - 3.0 ->  5Min
    (3.0, 900),  # 3.0 - 4.0 -> 15Min
    (4.0, 3600),  # > 4.0 ->  1h
))

# Sort from biggest to lowest:
SMOOTH_CACHE_TIMES = list(SMOOTH_CACHE_TIMES)
SMOOTH_CACHE_TIMES.sort(reverse=True)
SMOOTH_CACHE_TIMES = tuple(SMOOTH_CACHE_TIMES)


def get_max_age(load_average):
    """
    return max age for the given load average.

    >>> get_max_age(0)
    5
    >>> get_max_age(0.09)
    5
    >>> get_max_age(0.1)
    5
    >>> get_max_age(0.11)
    10
    """
    for load, max_age in SMOOTH_CACHE_TIMES:
        if load_average > load:
            break
    return max_age


class SmoothCacheTime(int):
    def __new__(cls, value=None):
        if value is None:
            value = time.time()
        i = int.__new__(cls, value)
        return i


class _SmoothCache:
    __CHANGE_TIME = None  # Timestamp of the "last update"
    __NEXT_SYNC = None  # Point in the future to update the __CHANGE_TIME

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._smooth_clear = None

    def __get_change_time(self):
        """
        return current "last change" timestamp.
        To save cache access, update the timestamp only in
        SMOOTH_CACHE_UPDATE_TIMESTAMP frequency from cache.
        """
        now = time.time()
        if now > self.__NEXT_SYNC or self.__CHANGE_TIME is None:
            # Update timestamp we must look for a new change time from cache:
            self.__NEXT_SYNC = now + SMOOTH_CACHE_UPDATE_TIMESTAMP

            # use raw method, otherwise: end in a endless-loop ;)
            change_time = self.get(SMOOTH_CACHE_CHANGE_TIME, raw=True)
            if change_time is None:
                logger.debug("CHANGE_TIME is None")
                self.smooth_update()  # save change time into cache
            elif change_time > self.__CHANGE_TIME:
                self.__CHANGE_TIME = change_time
                logger.debug(f"update change time to: {change_time!r}")
#        else:
#            logger.debug("Use old CHANGE_TIME %r" % self.__CHANGE_TIME)

        return self.__CHANGE_TIME

    def __must_updated(self, key, create_time):
        """
        return True if given cache create time is older than the "last change"
        time, but only if the additional time from system load allows it.
        """
        last_change_time = self.__get_change_time()
        if last_change_time < create_time:
            # Item was added to the cache after the last clear() time.
            logger.debug(
                f"{key!r} not out-dated: added {create_time - last_change_time}sec before clear()")
            return False

        outdate_age = time.time() - last_change_time
        load_average = os.getloadavg()[0]  # load over last minute
        max_age = get_max_age(load_average)
        if outdate_age > max_age:
            logger.debug(
                f"Out-dated {key!r}"
                f" (added {last_change_time - create_time}sec after clear()"
                f" - age: {outdate_age}, max age: {max_age}, load: {load_average})"
            )
            return True
        else:
            logger.debug(
                f"Keep {key!r} by load"
                f" (out-dated age: {outdate_age:.1f}sec, max age: {max_age}, load: {load_average})"
            )
            return False

    def smooth_update(self):
        """
        save the "last change" timestamp to renew the cache entries in
        self.__CHANGE_TIME and in cache.
        """
        now = int(time.time())
        self.__CHANGE_TIME = now
        self.set(SMOOTH_CACHE_CHANGE_TIME, now, raw=True)  # will be get via __origin_get()
        logger.debug(f"Set CHANGE_TIME to {now!r}")

    # --------------------------------------------------------------------------

    def get(self, key, default=None, version=None, raw=False):
        value = super().get(key, default, version)
        if raw:
            return value
        if value is None or value is default:
            # Item not in cache
            return value

        try:
            create_time, value = value
            assert isinstance(create_time, SmoothCacheTime), (
                f"create_time is not SmoothCacheTime instance, it's: {type(create_time)}"
            )
        except Exception as err:
            # e.g: entry is saved before smooth cache used.
            logger.error(f"Can't get 'create_time' from: {err} (Maybe {key!r} is a old cache entry?)")
            self.delete(key, version)
            return default

        if self.__must_updated(key, create_time):
            # is too old -> delete the item
            self.delete(key, version)
            return default

        return value

    def set(self, key, value, timeout=None, version=None, raw=False):
        if not raw:
            value = (SmoothCacheTime(), value)
        super().set(key, value, timeout, version)

    def clear(self):
        logger.debug("SmoothCache clear called!")
        super().clear()
        self.smooth_update()


class SmoothFileBasedCache(_SmoothCache, FileBasedCache):
    pass


class SmoothDatabaseCache(_SmoothCache, DatabaseCache):
    pass


class SmoothLocMemCache(_SmoothCache, LocMemCache):
    pass


class SmoothMemcachedCache(_SmoothCache, MemcachedCache):
    pass


class SmoothPyLibMCCache(_SmoothCache, PyLibMCCache):
    pass



"""
    Limit to usergroups
    ~~~~~~~~~~~~~~~~~~~

    Helper to limit something to:
        * for everyone (anonymous users)
        * for staff users
        * for superusers
        * for a user group

    In forms you will get a select field with:
        * anonymous users
        * staff users
        * superusers
        * ..all existing user groups..


    usage example
    ~~~~~~~~~~~~~

    model.py
    ---------------------------------------------------------------------------
    class FooBar(models.Model):
        permit_edit = limit_to_usergroups.UsergroupsModelField()
        permit_view = limit_to_usergroups.UsergroupsModelField()
    ---------------------------------------------------------------------------

    views.py
    ---------------------------------------------------------------------------
    def view(request):
        queryset = FooBar.objects.all()
        queryset = limit_to_usergroups.filter_permission(queryset, permit_view=request.user)
        ...

    def edit(request, id):
        foo = FooBar.objects.get(id=id)

        if not limit_to_usergroups.has_permission(poll, permit_edit=request.user):
            msg = _("You have no permission to edit this.")
            if settings.DEBUG:
                verbose_limit_name = limit_to_usergroups.get_verbose_limit_name(foo.permit_edit)
                msg += " (Limited to: %s)" % verbose_limit_name
            messages.error(request, msg)
            return HttpResponseRedirect("foobar)
        ...
    ---------------------------------------------------------------------------


    :copyleft: 2011-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.contrib.auth.models import Group
from django.db import models
from django.utils.translation import gettext_lazy as _


def has_permission(item, **kwargs):
    """
    return True/False if use has permission stored in UsergroupsModelField.

    return True, if permission are ok
    return False, if permission are not sufficient.

    e.g.:
        has_permission(poll, permit_vote=request.user)
        has_permission(item, model_field1=user, model_field2=user)
    """
    for attr_name, user in list(kwargs.items()):
        limit_permission_value = getattr(item, attr_name)
        limit_permission_value = int(limit_permission_value)

        if limit_permission_value == UsergroupsModelField.ANONYMOUS_USERS:
            continue

        if user.is_anonymous:
            return False

        if limit_permission_value == UsergroupsModelField.NORMAL_USERS:
            if user.is_authenticated:
                continue
            else:
                return False

        if limit_permission_value == UsergroupsModelField.STAFF_USERS:
            if user.is_staff:
                continue
            else:
                return False

        if limit_permission_value == UsergroupsModelField.SUPERUSERS:
            if user.is_superuser:
                continue
            else:
                return False

        usergroup = Group.objects.get(id=limit_permission_value)
        usergroups = user.groups.all()
        if usergroup not in usergroups:
            return False

    return True


def filter_permission(queryset, **kwargs):
    """
    Filter a given queryset with UsergroupsModelField.

    e.g.:
        queryset = filter_permission(queryset, permit_view=request.user)
        queryset = filter_permission(queryset, model_field1=user, model_field2=user)

    FIXME: Would be nice, if this is a normal QuerySet method
    """
    result = [item for item in queryset if has_permission(item, **kwargs)]
    return result


def get_verbose_limit_name(value):
    """
    Simply convert the integer value of a UsergroupsModelField to his select choice text

    e.g.:
        x = get_verbose_limit_name(-1)
        print(x) # "staff users"
    """
    limit_dict = get_limit_dict()
    return limit_dict[value]


def get_user_groups():
    groups = Group.objects.values_list("id", "name")
    return list(groups)


def get_limit_dict():
    # use unicode() to evaluate gettext_lazy:
    limit_dict = {k: str(v) for k, v in list(UsergroupsModelField.USER_TYPES_DICT.items())}

    groups = get_user_groups()
    limit_dict.update(dict(groups))
    return limit_dict


class UsergroupsModelField(models.IntegerField):
    """
    TODO: Use html select optgroup [1] to group anonymous, staff and superusers
    from user groups
        [1] http://www.w3.org/wiki/HTML/Elements/optgroup
    """
    ANONYMOUS_USERS = 0
    STAFF_USERS = -1
    SUPERUSERS = -2
    NORMAL_USERS = -3

    USER_TYPES_CHOICES = [
        (ANONYMOUS_USERS, _("anonymous users")),
        (NORMAL_USERS, _("normal users")),
        (STAFF_USERS, _("staff users")),
        (SUPERUSERS, _("superusers")),
    ]
    USER_TYPES_DICT = dict(USER_TYPES_CHOICES)

    def __init__(self, *args, **kwargs):
        """
        We added only the static choices here, because we should not access
        the database here (can break unittests)
        In this cases the choices feature of this field is activated, but
        in any case self.get_choices() would be called: There we append
        the user groups
        """
        kwargs["choices"] = self.USER_TYPES_CHOICES

        super().__init__(*args, **kwargs)

    def get_choices(self, *args, **kwargs):
        """
        Built the choices list with the static part and all existing
        user groups from database.
        """
        groups = get_user_groups()
        choices = self.USER_TYPES_CHOICES + list(groups)
        return choices



"""
    models stuff
    ~~~~~~~~~~~~

    :copyleft: 2011-2019 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.timezone import now

# https://github.com/jedie/django-tools
from django_tools.middlewares import ThreadLocal


class UpdateTimeBaseModel(models.Model):
    """
    Base model to automatically set:
     * createtime
     * lastupdatetime

    We doesn't used auto_now_add and auto_now here, because they have the odd side effect
    see also:
    https://github.com/jezdez/django-dbtemplates/commit/2f27327bebe7f2e7b33e5cfb0db517f53a1b9701#commitcomment-1396126
    """

    createtime = models.DateTimeField(default=now, editable=False, help_text="Create time")
    lastupdatetime = models.DateTimeField(default=now, editable=False, help_text="Time of the last change.")

    def __str__(self):  # to be overwritten
        return f"model instance ID:{self.pk}"

    def save(self, *args, **kwargs):
        self.lastupdatetime = now()
        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


class UpdateUserBaseModel(models.Model):
    """
    Base model to automatically set:
     * createby
     * lastupdateby

    Important: "threadlocals middleware" must be used!
    """

    createby = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        related_name="%(class)s_createby",
        null=True,
        blank=True,  # <- If the model used outside a real request (e.g. unittest, db shell)
        help_text="User how create this entry.",
        on_delete=models.SET_NULL,
    )
    lastupdateby = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        related_name="%(class)s_lastupdateby",
        null=True,
        blank=True,  # <- If the model used outside a real request (e.g. unittest, db shell)
        help_text="User as last edit this entry.",
        on_delete=models.SET_NULL,
    )

    def __str__(self):  # to be overwritten
        return f"model instance ID:{self.pk}"

    def save(self, *args, **kwargs):
        current_user = ThreadLocal.get_current_user()

        if current_user:
            User = get_user_model()
            if isinstance(current_user, User):
                if self.pk is None or kwargs.get("force_insert", False):  # New model entry
                    self.createby = current_user
                self.lastupdateby = current_user

        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


class UpdateInfoBaseModel(UpdateTimeBaseModel, UpdateUserBaseModel):
    """
    Base model to automatically set:
        * createtime
        * lastupdatetime
    and:
        * createby
        * lastupdateby

    Important: "threadlocals middleware" must be used!
    """

    class Meta:
        abstract = True



"""
    :created: 24.04.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.core.management import BaseCommand
from django.utils import translation

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.model_test_code_generator import ModelTestGenerator


class Command(BaseCommand):
    """
    List all models, e.g.:
        $ ./manage.py generate_model_test_code

    Generate test code, e.g.:
        $ ./manage.py generate_model_test_code auth
        $ ./manage.py generate_model_test_code auth.User
        $ ./manage.py generate_model_test_code cms.

    """
    help = "Generate unittest code for a model"

    def add_arguments(self, parser):
        parser.add_argument('model_label', nargs="?",
                            help='Name of the Django model (can only be the first characters! We use "startwith"'
                            )
        parser.add_argument('--translation', help='Language code that will be activated', default="en")
        parser.add_argument('--count', help='Number of data records to be generated.', type=int, default=2)

    def handle(self, *args, **options):
        model_label = options['model_label']
        language_code = options["translation"]
        count = int(options["count"])

        translation.activate(language_code)

        model_test_generator = ModelTestGenerator()

        if model_label is None:
            print("\nNo model_label given.\n")
            model_test_generator.print_all_plugins()
            return

        models = model_test_generator.get_models_startwith_label(model_label)
        if not models:
            print(f"\nERROR: No models starts with given label {model_label!r}\n")
            model_test_generator.print_all_plugins()
            return

        for model in models:
            queryset = model.objects.all()

            existing_count = queryset.count()
            if existing_count == 0:
                print("#")
                print(f"# ERROR: {model._meta.label!r} is empty!")
                print("#")
                continue

            if existing_count < count:
                print("#")
                print(f"# Warning: They exists only {existing_count:d} items in {model._meta.label!r}!")
                print("#")

            queryset = queryset[:count]

            content = model_test_generator.from_queryset(queryset)
            print(content)



"""
    :created: 2015 by Jens Diemer
    :copyleft: 2015-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging
import time
import warnings

from django.http import HttpResponse, SimpleCookie
from django.test import RequestFactory


log = logging.getLogger(__name__)


class FakedHttpResponse(HttpResponse):
    """
    Used in selenium tests.
    So django assert statements like
    assertContains() can be used.
    """

    pass


def selenium2faked_response(driver, client, client_class):
    """
    Create a similar 'testing-response' [1] here.
    So that some of the django testing assertions [2] can be used
    with selenium tests, too ;)

    [1] https://docs.djangoproject.com/en/1.7/topics/testing/tools/#testing-responses
    [2] https://docs.djangoproject.com/en/1.7/topics/testing/tools/#assertions

    Currently not available:
        * response.status_code
        * response.redirect_chain
        * response.templates
        * response.context

    Available:
        * response.content
        * response.cookies
        * response.client.cookies
        * response.session

    usage e.g.:
        from django_tools.unittest_utils.selenium_utils import selenium2faked_response

        class MySeleniumTests(StaticLiveServerTestCase):
            def get_faked_response(self):
                return selenium2faked_response(self.driver, self.client, self.client_class)
            def test_foo(self):
                self.driver.get("/foo")
                faked_response = self.get_faked_response()
                self.assertNotContains(response, "<h1>foobar</h1>", html=True)
    """
    response = FakedHttpResponse(content=driver.page_source)
    response.client = client_class()  # Fresh Client() instance

    # Add 'response.client.cookies':
    # driver.get_cookies() is a simple list of dict items, e.g.:
    # [{'name': 'csrftoken', 'value': 'yXoN3...', ...},...]
    cookies = SimpleCookie()
    for cookie in driver.get_cookies():
        key = cookie.pop("name")
        cookies[key] = cookie.pop("value")
        for k, v in cookie.items():
            if k == "expiry":
                cookies[key]["expires"] = time.time() - v
            else:
                cookies[key][k] = v

    # response.cookies and response.client.cookies
    response.cookies = response.client.cookies = cookies
    # print("\nresponse.cookies:", response.cookies)

    # Add 'response.session':
    response.session = response.client.session
    # print("\nresponse.session:", dict(response.session))

    help(driver)

    response.request = RequestFactory()
    response.request.path = driver.current_url

    return response


def selenium2fakes_response(*args, **kwargs):
    warnings.warn(
        "selenium2fakes_response() is deprecated, use selenium2faked_response() !", category=DeprecationWarning
    )
    return selenium2faked_response(*args, **kwargs)



"""
    :created: 2015 by Jens Diemer
    :copyleft: 2015-2020 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging
import shutil

from selenium import webdriver
from selenium.webdriver import DesiredCapabilities

# https://github.com/jedie/django-tools
from django_tools.selenium.base import SeleniumBaseTestCase


log = logging.getLogger(__name__)


class SeleniumChromiumTestCase(SeleniumBaseTestCase):
    """
    TestCase with Selenium and the Chromium WebDriver
    Note:
        * Needs 'chromium-chromedriver' executable! See README.creole for more info
        * It's without django StaticLiveServerTestCase

    usage e.g.:

        from django_tools.selenium.chromedriver import SeleniumChromiumTestCase, chromium_available

        @unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
        class ChromiumTests(SeleniumChromiumTestCase):

            def test_admin_login_page(self):
                self.driver.get(self.live_server_url + "/admin/login/")
                self.assert_equal_page_title("Log in | Django site admin")
                self.assert_in_page_source(
                    '<form action="/admin/login/" method="post" id="login-form">'
                )
                self.assert_no_javascript_alert()

    see also: django_tools_tests/test_unittest_selenium.py
    """
    filename = "chromedriver"

    options = (
        "--no-sandbox",
        # "--headless", ## see: https://bugs.chromium.org/p/chromedriver/issues/detail?id=3358
        "--disable-gpu",
        "--disable-dev-shm-usage",  # https://bugs.chromium.org/p/chromedriver/issues/detail?id=2473
    )
    desired_capabilities = {
        "loggingPrefs": {
            "browser": "ALL",
            "client": "ALL",
            "driver": "ALL",
            "performance": "ALL",
            "server": "ALL"
        }
    }
    accept_languages = 'en'

    @classmethod
    def setUpClass(cls):

        options = webdriver.ChromeOptions()
        options.add_experimental_option('w3c', False)  # needed to get browser logs
        options.add_experimental_option('prefs', {'intl.accept_languages': cls.accept_languages})

        for argument in cls.options:
            options.add_argument(argument)

        try:
            executable = shutil.which(cls.filename)
        except FileNotFoundError:
            cls.driver = None
        else:
            desired = DesiredCapabilities.CHROME
            for key, value in cls.desired_capabilities.items():
                desired[key] = value

            cls.driver = webdriver.Chrome(
                options=options,
                executable_path=str(executable),  # Path() instance -> str()
                desired_capabilities=desired,
            )
            cls.driver.implicitly_wait(10)

        super().setUpClass()

    def get_browser_log(self):
        assert "browser" in self.driver.log_types
        browser_log = self.driver.get_log("browser")
        lines = []
        for entry in browser_log:
            lines.append(("{timestamp} {level} {source} {message}").format(**entry))
        return "\n".join(lines)

    def assert_in_browser_log(self, text):
        log = self.get_browser_log()
        print(log)
        self.assertIn(text, log)


def chromium_available(filename=None):
    """
    :return: True/False if 'chromium-chromedriver' executable can be found

    >>> chromium_available("doesn't exists")
    False
    """
    if filename is None:
        filename = SeleniumChromiumTestCase.filename

    executable = shutil.which(filename)
    if not executable:
        log.error("Chromium is no available!")
        return False

    log.debug(f"Chromium found here: {executable}")
    return True



"""
    :created: 28.08.2018 by Jens Diemer
    :copyleft: 2018-2020 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import textwrap
from pathlib import Path
from warnings import WarningMessage

import icdiff
import pprintpp
from django.conf import settings
from django.core import mail
from django.core.mail import get_connection


def assert_startswith(text, prefix):
    """
    Check if test starts with prefix.
    """
    assert text.startswith(prefix), f"{text!r} doesn't starts with {prefix!r}"


def assert_endswith(text, suffix):
    """
    Check if text ends with suffix.
    """
    assert text.endswith(suffix), f"{text!r} doesn't ends with {suffix!r}"


def assert_locmem_mail_backend():
    """
    Check if current email backend is the In-memory backend.
    See:
        https://docs.djangoproject.com/en/1.11/topics/email/#in-memory-backend
    """
    mail_backend = get_connection()
    assert isinstance(mail_backend, mail.backends.locmem.EmailBackend), f"Wrong backend: {mail_backend}"


def assert_language_code(*, language_code):
    """
    Check if given language_code is in settings.LANGUAGES
    """
    existing_language_codes = tuple(dict(settings.LANGUAGES).keys())
    assert language_code in existing_language_codes, (
        f"{language_code!r} not in settings.LANGUAGES={settings.LANGUAGES!r}"
    )


def assert_installed_apps(*, app_names):
    """
    Check entries in settings.INSTALLED_APPS
    """
    assert isinstance(app_names, (tuple, list))
    installed_apps = settings.INSTALLED_APPS
    for app_name in app_names:
        assert app_name in installed_apps, f"{app_name!r} not in settings.INSTALLED_APPS!"


def assert_is_dir(path):
    """
    Check if given path is a directory
    """
    if not isinstance(path, Path):
        path = Path(path)
    assert path.is_dir(), f"Directory not exists: {path}"


def assert_is_file(path):
    """
    Check if given path is a file
    """
    if not isinstance(path, Path):
        path = Path(path)
    assert_is_dir(path.parent)
    assert path.is_file(), f"File not exists: {path}"


def assert_path_not_exists(path):
    """
    Check if given path doesn't exists
    """
    if not isinstance(path, Path):
        path = Path(path)

    assert not path.is_dir(), f"Path is a existing directory: {path}"
    assert not path.is_file(), f"Path is a existing file: {path}"
    assert not path.is_fifo(), f"Path is a existing fifo: {path}"
    assert not path.exists(), f"Path exists: {path}"


def pformat(obj, indent=4, width=100):
    pformat_return_list = pprintpp.pformat(obj, indent=indent, width=width)
    return pformat_return_list.splitlines()


def create_icdiff(first, second, fromfile="first", tofile="second", indent=4, width=100):
    """
    Based on:
        https://github.com/hjwp/pytest-icdiff/blob/master/pytest_icdiff.py
    """
    if isinstance(first, str):
        first = first.splitlines()
    else:
        first = pformat(first, indent=indent, width=width)

    if isinstance(second, str):
        second = second.splitlines()
    else:
        second = pformat(second, indent=indent, width=width)

    icdiff_lines = list(
        icdiff.ConsoleDiff(tabsize=2, cols=width, highlight=True).make_table(
            fromlines=first, tolines=second, fromdesc=fromfile, todesc=tofile
        )
    )
    if len(icdiff_lines) == 1:
        # hacky whitespace reduce:
        icdiff_lines[0] = icdiff_lines[0].replace("        ", " ")

    # icdiff_lines = [f"{COLOR_OFF}{l}" for l in icdiff_lines]

    return "\n".join(icdiff_lines)


def assert_pformat_equal(first, second, msg="", **pformat_kwargs):
    """ compare with pprintpp and icdiff output """
    if first != second:
        if isinstance(first, str):
            print(first)
        else:
            pprintpp.pprint(first, **pformat_kwargs)
        assert first == second, f"{msg}{create_icdiff(first=first, second=second, **pformat_kwargs)}"


def dedent(txt):
    # Remove any common leading whitespace from every line
    txt = textwrap.dedent(txt)

    # strip whitespace at the end of every line
    txt = "\n".join([line.rstrip() for line in txt.splitlines()])
    txt = txt.strip()
    return txt


def assert_equal_dedent(first, second, msg=""):
    assert_pformat_equal(dedent(first), dedent(second), msg=msg)


def assert_in_dedent(member, container):
    member = dedent(member)
    container = dedent(container)
    assert member in container, f"{member!r} not found in {container!r}"


def assert_filenames_and_content(*, path, reference, fromfile="current", tofile="reference", **pformat_kwargs):
    if not isinstance(path, Path):
        path = Path(path)

    assert_is_dir(path)

    current_data = []
    for item in sorted(path.iterdir()):
        with item.open("rb") as f:
            current_data.append((item.name, f.read()))

    if current_data != reference:
        print("\nCurrent filenames and content:")
        pprintpp.pprint(current_data, **pformat_kwargs)

    assert_pformat_equal(current_data, reference, fromfile=fromfile, tofile=tofile, **pformat_kwargs)


def assert_warnings(messages, reference):
    """
    e.g.:
        with warnings.catch_warnings(record=True) as w:
            warnings.warn('test')
            assert_warnings(
                messages=w,
                reference=["UserWarning('test')"]
            )
    """
    msg_text = []
    for message in messages:
        assert isinstance(message, WarningMessage)
        msg_text.append(repr(message.message))

    assert_pformat_equal(msg_text, reference)


def assert_no_warnings(messages):
    """
    Check that there are no warning created.
    e.g.:
        with warnings.catch_warnings(record=True) as w:
            # ...do something...
            assert_no_warnings(w)
    """
    if not messages:
        return

    msg_text = []
    for message in messages:
        assert isinstance(message, WarningMessage)
        msg_text.append(str(message))

    raise AssertionError(msg_text)


def assert_in_logs(logs, line):
    """
    e.g.:
        with self.assertLogs(logger="django_tools.local_sync_cache") as logs:
            # ...do something...
            assert_in_logs(logs, line="ERROR:foo.bar:Foo Bar!")
    """
    output = [str(entry) for entry in logs.output]
    assert line in output



"""
    unittests for filemanager
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2012-2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import os
import tempfile
import unittest

from django.http import Http404

from django_tools.filemanager.exceptions import DirectoryTraversalAttack
from django_tools.filemanager.filemanager import BaseFilemanager


class FilemanagerBaseTestCase(unittest.TestCase):
    """
    Create this test filesystem tree:

    self.BASE_PATH/file1.txt
    self.BASE_PATH/file2.txt
    self.BASE_PATH/subdir1/subfile1.txt
    self.BASE_PATH/subdir1/subsubdir1/
    self.BASE_PATH/subdir1/subsubdir1/subsubfile1.txt
    self.BASE_PATH/subdir1/subsubdir1/subsubfile2.txt
    self.BASE_PATH/subdir1/emptysubsubdir/
    self.BASE_PATH/emptysubdir/
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.BASE_PATH = f"{tempfile.mkdtemp(prefix='filemanager-unittests_')}/"

        # For dir traversal attack tests
        cls.SUB_BASE_PATH = os.path.join(cls.BASE_PATH, "subdir1")
        cls.ATTACK_PARTS = ("..", "..\\", "../",
                            # from
                            # https://en.wikipedia.org/wiki/Directory_traversal_attack#URI_encoded_directory_traversal
                            "%2e%2e%2f", "%2e%2e/", "..%2f", "%2e%2e%5c"
                            # from
                            # https://en.wikipedia.org/wiki/Directory_traversal_attack#Unicode_.2F_UTF-8_encoded_directory_traversal
                            "..%c1%1c", "..%c0%af",
                            "..\xc1\x1c", "..\xc0\xaf",
                            )

        cls.DIRS = (
            (cls.BASE_PATH, "subdir1", "subsubdir1"),
            (cls.BASE_PATH, "subdir1", "emptysubsubdir"),
            (cls.BASE_PATH, "emptysubdir"),
        )
        cls.FILES = (
            (cls.BASE_PATH, "file1.txt"),
            (cls.BASE_PATH, "file2.txt"),
            (cls.BASE_PATH, "subdir1", "subfile1.txt"),
            (cls.BASE_PATH, "subdir1", "subsubdir1", "subsubfile1.txt"),
            (cls.BASE_PATH, "subdir1", "subsubdir1", "subsubfile2.txt"),
        )

        # Create dirs:
        for dirs in cls.DIRS:
            path = os.path.join(*dirs)
#            print "Create dir: %s" % repr(path)
            os.makedirs(path)

        # Create files:
        for file_info in cls.FILES:
            path = os.path.join(*file_info)
            content = f"File content for: {file_info[-1]!r}"
#            print "Create file %s with %s" % (repr(path), repr(content))
            with open(path, "w") as f:
                f.write(content)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        tempdir = tempfile.gettempdir()
        if cls.BASE_PATH.startswith(tempdir):
            for root, dirs, files in os.walk(cls.BASE_PATH, topdown=False):
                for name in files:
                    path = os.path.join(root, name)
                    os.remove(path)
                for name in dirs:
                    path = os.path.join(root, name)
                    os.rmdir(path)
            os.rmdir(cls.BASE_PATH)
        else:
            cls.fail(f"Cleanup error: {cls.BASE_PATH} not in {tempdir}")

    def test_setup(self):
        """ if this tests fails, all other may fail, too. """
        for file_info in self.FILES:
            path = os.path.join(*file_info)
            self.assertTrue(os.path.isfile(path))


class FilemanagerDirectoryTraversal(FilemanagerBaseTestCase):

    def test_subdir1(self):
        subdir = "subdir1"
        base_url = "/base/url/"
        rest_url = f"{subdir}/"
        fm = BaseFilemanager(None, self.BASE_PATH, base_url, rest_url)
        self.assertEqual(fm.abs_url, '/base/url/subdir1/')
        self.assertEqual(fm.absolute_path, self.BASE_PATH)
        self.assertEqual(fm.breadcrumbs, [
            {'url': base_url, 'name': 'index', 'title': "goto 'index'"},
            {'url': f'/base/url/{subdir}/', 'name': subdir, 'title': f"goto '{subdir}'"}
        ])

    def test_not_existing_path1(self):
        self.assertRaises(Http404,
                          BaseFilemanager, None, self.BASE_PATH, "/base/url/", "not_exists"
                          )

    def test_not_existing_path2(self):
        self.assertRaises(Http404,
                          BaseFilemanager, None, self.BASE_PATH, "/base/url/", "does/not/exists"
                          )

    def test_dir_traversal_attack1(self):
        for parts in self.ATTACK_PARTS:
            self.assertRaises((DirectoryTraversalAttack, Http404),
                              BaseFilemanager, None, self.SUB_BASE_PATH, "/base/url/", parts,
                              )

    def test_dir_traversal_attack2(self):
        for parts in self.ATTACK_PARTS:
            self.assertRaises((DirectoryTraversalAttack, Http404),
                              BaseFilemanager, None, self.SUB_BASE_PATH, "/base/url/", f"{parts}emptysubdir",
                              )

    def test_dir_traversal_attack3(self):
        self.assertRaises(Http404,
                          BaseFilemanager, None, self.BASE_PATH, "/base/url/", "subdir1/../../etc/passwd"
                          )

    def test_dir_traversal_attack4(self):
        self.assertRaises(Http404,
                          BaseFilemanager, None, self.BASE_PATH, "/base/url/", "subdir1/c:\\boot.ini"
                          )



from importlib import import_module

from django.apps import apps
from django.urls import Resolver404
from django.urls.resolvers import RegexPattern, URLResolver


def get_filtered_apps(resolve_url="/", no_args=True, debug=False, skip_fail=False):
    """
    Create a list of all Apps witch can resolve the given url >resolve_url<

    @param resolve_url: url used for RegexURLResolver
    @param no_args: Only views without args/kwargs ?
    @parm skip_fail: If True: raise exception if app is not importable

    Please look at:

        django_tools.django_tools_tests.test_installed_apps_utils

    with debug, some print messages would be created:

    e.g.: get_filtered_apps(debug=True)
    found 'django.contrib.admindocs' with urls.py
    found 'django.contrib.auth' with urls.py
    Skip 'django.contrib.auth': Can't handle root url.
    found 'django.contrib.flatpages' with urls.py
    ['django.contrib.admindocs']
    """
    root_apps = []
    app_configs = apps.get_app_configs()
    for app_config in app_configs:
        app_pkg = app_config.name
        urls_pkg = f"{app_pkg}.urls"
        try:
            url_mod = import_module(urls_pkg)
        except ImportError as err:
            if debug:
                print(f"Skip {app_pkg!r}: has no urls.py")
            if str(err) == f"No module named '{urls_pkg}'":
                continue
            if not skip_fail:
                raise
        except Exception as err:
            if debug:
                print(f"Error importing {app_pkg!r}: {err}")
            if not skip_fail:
                raise
            else:
                continue

        if debug:
            print(f"found {app_pkg!r} with urls.py")

        try:
            urlpatterns = url_mod.urlpatterns
        except AttributeError:
            if debug:
                print(f"Skip {app_pkg!r}: urls.py has no 'urlpatterns'")
            continue

        resolver = URLResolver(RegexPattern(r'^'), urlpatterns)
        try:
            func, func_args, func_kwargs = resolver.resolve(resolve_url)
        except Resolver404 as err:
            if debug:
                print(f"Skip {app_pkg!r}: Can't handle root url. ({err})")
            continue
        if not no_args or func_args == () and func_kwargs == {}:
            root_apps.append(app_pkg)
    return root_apps



"""
    utils for filemanager
    ~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import ntpath
import os
import posixpath


def clean_posixpath(path, up_level_references=False):
    """
    Based on code from:
    https://github.com/django/django/blob/master/django/views/static.py

    But we build always a posixpath with "/" as path separate character.


    Keep starting/ending slash:

    >>> clean_posixpath("no/slash")
    'no/slash'
    >>> clean_posixpath("/starts/with/slash")
    '/starts/with/slash'
    >>> clean_posixpath("ends/with/slash/")
    'ends/with/slash/'
    >>> clean_posixpath("") # normpath would return: "."
    ''
    >>> clean_posixpath("/")
    '/'
    >>> clean_posixpath("/../")
    '/'


    Remove every crude characters:

    >>> clean_posixpath("foo//bar")
    'foo/bar'
    >>> clean_posixpath("/foo/./bar")
    '/foo/bar'
    >>> clean_posixpath(r"foo\\bar/")
    'foo/bar/'


    up-level references would be only applied if activated:

    >>> clean_posixpath("/foo/bar/../../etc/passwd") # normpath would return: '../etc/passwd'
    '/foo/bar/etc/passwd'
    >>> clean_posixpath("/foo/bar/../../etc/passwd", up_level_references=True)
    '/etc/passwd'

    >>> clean_posixpath("../../../etc/passwd") # normpath would return: '../../../etc/passwd'
    'etc/passwd'

    >>> clean_posixpath(r"\\foo\\bar\\..\\etc\\passwd") # normpath would return: '\\foo\\bar\\..\\etc\\password'
    '/foo/bar/etc/passwd'


    Ignore windows drive parts:

    >>> clean_posixpath(r"c:\\boot.ini")
    'boot.ini'
    >>> clean_posixpath(r"foo/bar/c:\\boot.ini")
    'foo/bar/boot.ini'
    """
    path = path.replace('\\', '/')

    add_slash = path.endswith("/")
    if path.startswith("/"):
        newpath = "/"
    else:
        newpath = ""
    path = path.strip("/")

    if up_level_references:
        # e.g.: foo/../bar -> bar
        path = posixpath.normpath(path)

    for part in path.split("/"):
        if not part:
            # Strip empty path components.
            continue

        drive, part = ntpath.splitdrive(part)
        head, part = posixpath.split(part)
        if part in (".", ".."):
            continue
        newpath = posixpath.join(newpath, part)

    if add_slash and newpath != "/":
        newpath += "/"

    return newpath


def add_slash(path):
    """
    >>> add_slash("/foo")
    '/foo/'
    >>> add_slash("/bar/")
    '/bar/'
    """
    if not path.endswith(os.sep):
        path += os.sep
    return path


def symbolic_notation(mode):
    """
    Convert os.stat().st_mode values to a symbolic representation. e.g:

    >>> s = symbolic_notation(16893) # -> 040775 -> 775
    >>> s == 'rwxrwxr-x'
    True

    >>> s = symbolic_notation(33204) # -> 0100664 -> 664
    >>> s == 'rw-rw-r--'
    True
    """
    mode = mode & 0o777  # strip "meta info"
    chmod_symbol = ''.join(
        mode & 0o400 >> i and x or '-' for i, x in enumerate('rwxrwxrwx')
    )
    return chmod_symbol



"""
    SlowerDevServer
    ~~~~~~~~~~~~~~~

    Simple slow down the django developer server.
    The middleware insert in every 200 response a time.sleep

    Put this into your settings:
    --------------------------------------------------------------------------
        MIDDLEWARE_CLASSES = (
            ...
            'django_tools.middlewares.SlowerDevServer.SlowerDevServerMiddleware',
            ...
        )
        SLOWER_DEV_SERVER_SLEEP = 0.3 # time.sleep() value (in sec.)
    --------------------------------------------------------------------------

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author:$

    :copyleft: 2009 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import time
import warnings

from django.conf import settings


try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object  # fallback for Django < 1.10


class SlowerDevServerMiddleware(MiddlewareMixin):
    def __init__(self):
        warnings.warn("Slower developer server used!")

    def process_response(self, request, response):
        if response.status_code == 200:
            print(f"SlowerDevServerMiddleware: Wait for {settings.SLOWER_DEV_SERVER_SLEEP}Sec...")
            time.sleep(settings.SLOWER_DEV_SERVER_SLEEP)
        return response
